"""
Unit tests for plots.py functions.
"""

import unittest
import matplotlib
import matplotlib.pyplot as plt
from radioactivedecay.plots import (
    _parse_nuclide_label,
    _parse_decay_mode_label,
    _check_fig_axes,
)


class TestPlotsFunctions(unittest.TestCase):
    """
    Unit tests for plots.py functions.
    """

    def test__parse_nuclide_label(self) -> None:
        """
        Test the parsing of nuclide strings for node labels.
        """

        self.assertEqual(_parse_nuclide_label("H-3"), "³H")
        self.assertEqual(_parse_nuclide_label("Be-7"), "⁷Be")
        self.assertEqual(_parse_nuclide_label("C-10"), "¹⁰C")
        self.assertEqual(_parse_nuclide_label("Ne-19"), "¹⁹Ne")
        self.assertEqual(_parse_nuclide_label("I-118"), "¹¹⁸I")
        self.assertEqual(_parse_nuclide_label("Pd-100"), "¹⁰⁰Pd")
        self.assertEqual(_parse_nuclide_label("Cl-34m"), "³⁴ᵐCl")
        self.assertEqual(_parse_nuclide_label("I-118m"), "¹¹⁸ᵐI")
        self.assertEqual(_parse_nuclide_label("Tb-156m"), "¹⁵⁶ᵐTb")
        self.assertEqual(_parse_nuclide_label("Tb-156n"), "¹⁵⁶ⁿTb")
        self.assertEqual(_parse_nuclide_label("SF"), "various")

    def test__parse_decay_mode_label(self) -> None:
        """
        Test the parsing of decay mode strings for edge labels.
        """

        self.assertEqual(_parse_decay_mode_label("α"), "α")
        self.assertEqual(_parse_decay_mode_label("β+"), "β⁺")
        self.assertEqual(_parse_decay_mode_label("β+ & EC"), "β⁺ & EC")
        self.assertEqual(_parse_decay_mode_label("β-"), "β⁻")
        self.assertEqual(_parse_decay_mode_label("EC"), "EC")
        self.assertEqual(_parse_decay_mode_label("IT"), "IT")
        self.assertEqual(_parse_decay_mode_label("SF"), "SF")

    def test__check_fig_axes(self) -> None:
        """
        Test the parsing of user-defined Matplotlib Figure and Axes objects.
        """

        fig_in, axes_in = plt.subplots()
        fig, axes = _check_fig_axes(fig_in, axes_in)
        self.assertIsInstance(fig, matplotlib.figure.Figure)
        self.assertIsInstance(axes, matplotlib.axes.Axes)

        fig, axes = _check_fig_axes(fig_in, None)
        self.assertIsInstance(fig, matplotlib.figure.Figure)
        self.assertIsInstance(axes, matplotlib.axes.Axes)

        fig, axes = _check_fig_axes(None, axes_in)
        self.assertIsInstance(fig, matplotlib.figure.Figure)
        self.assertIsInstance(axes, matplotlib.axes.Axes)

        fig, axes = _check_fig_axes(None, None)
        self.assertIsInstance(fig, matplotlib.figure.Figure)
        self.assertIsInstance(axes, matplotlib.axes.Axes)


if __name__ == "__main__":
    unittest.main()



"""
The decaydata module defines the ``DecayData`` and ``DecayMatrices`` classes. Instances of
``DecayData`` initialize by reading in files containing radioactive decay data. The instances then
store the decay data, and their methods can be used for basic querying of the decay data.

The code examples shown in the docstrings assume the ``radioactivedecay`` package has been imported
as:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

Attributes
----------
DEFAULTDATA : DecayData
    Default radioactive decay dataset used by ``radioactivedecay``. This is currently ICRP-107.
"""

from pathlib import Path
import pickle
from typing import Any, ContextManager, Optional, Union
import numpy as np
from scipy import sparse
from sympy import log, Matrix
from sympy.matrices import SparseMatrix
from radioactivedecay.converters import (
    UnitConverterFloat,
    UnitConverterSympy,
    QuantityConverter,
    QuantityConverterSympy,
)
from radioactivedecay.utils import parse_nuclide

try:
    from importlib import resources
except ImportError:
    import importlib_resources as resources


def _get_package_filepath(subpackage_dir: str, filename: str) -> ContextManager[Path]:
    """
    Returns the path to a file which is bundled as a sub-package within the
    ``radioactivedecay`` package.

    Parameters
    ----------
    subpackage_dir : str
        Name of the sub-package directory.
    filename : str
        Name of the file.

    Returns
    -------
    ContextManager[Path]
        A context manager providing a file path object for the decay dataset file.
    """

    with resources.path(__package__ + "." + subpackage_dir, filename) as package_path:
        return package_path


def _load_package_pickle_file(subpackage_dir: str, filename: str) -> Any:
    """
    Returns an object loaded from a pickle file which is bundled as a sub-package within the
    ``radioactivedecay`` package.

    Parameters
    ----------
    subpackage_dir : str
        Name of the sub-package directory.
    filename : str
        Name of pickle file.

    Returns
    -------
    Object
        Object loaded from the pickle file.
    """

    with resources.open_binary(__package__ + "." + subpackage_dir, filename) as file:
        return pickle.load(file)


def _csr_matrix_equal(matrix_a: sparse.csr_matrix, matrix_b: sparse.csr_matrix) -> bool:
    """
    Checks whether two SciPy Compressed Sparse Row (CSR) matrices are equal (i.e. they have the
    same elements and all elements are equal).

    Parameters
    ----------
    matrix_a : scipy.sparse.csr.csr_matrix
        First SciPy CSR matrix.
    matrix_b : scipy.sparse.csr.csr_matrix
        Second SciPy CSR matrix.

    Returns
    -------
    bool
        True if both sparse matrices are equal, False otherwise.
    """

    return (
        np.array_equal(matrix_a.indptr, matrix_b.indptr)
        and np.array_equal(matrix_a.indices, matrix_b.indices)
        and np.array_equal(matrix_a.data, matrix_b.data)
    )


class DecayMatrices:
    """
    Instances of DecayMatrices store vectors with fundamental decay/atomic data and matrices used
    for decay calculations.

    Parameters
    ----------
    atomic_masses : numpy.ndarray
        Column vector of the atomic masses (in g/mol).
    decay_consts : numpy.ndarray
        Column vector of the decay constants (in s\\ :sup:`-1`).
    matrix_c : scipy.sparse.csr.csr_matrix
        A pre-calculated sparse lower traingular matrix used in decay calculations.
    matrix_c_inv : scipy.sparse.csr.csr_matrix
        The inverse of matrix_c, also used in decay calculations.

    Attributes
    ----------
    atomic_masses : numpy.ndarray
        Column vector of the atomic masses (in g/mol).
    decay_consts : numpy.ndarray
        Column vector of the decay constants (in s\\ :sup:`-1`).
    ln2: float
        Constant natural logarithm of 2.
    matrix_c : scipy.sparse.csr.csr_matrix
        A precalculated sparse lower triangular matrix used in decay calculations.
    matrix_c_inv : scipy.sparse.csr.csr_matrix
        The inverse of matrix_c, also used in decay calculations.
    matrix_e : scipy.sparse.csr.csr_matrix
        The matrix exponential that is used in radioactive decay calculations. It is a diagonal
        matrix that is pre-allocted for performance reasons.
    vector_n0 : numpy.ndarray
        Column vector for the number of atoms of each nuclide. It is pre-allocted here to improve
        the performance of decay calculations.

    """

    def __init__(
        self,
        atomic_masses: np.ndarray,
        decay_consts: np.ndarray,
        matrix_c: sparse.csr_matrix,
        matrix_c_inv: sparse.csr_matrix,
    ) -> None:
        self.atomic_masses = atomic_masses
        self.decay_consts = decay_consts
        self.ln2 = self._setup_ln2()
        self.matrix_c = matrix_c
        self.matrix_c_inv = matrix_c_inv
        self.matrix_e = self._setup_matrix_e(matrix_c.shape[0])
        self.vector_n0 = self._setup_vector_n0(matrix_c.shape[0])

    @staticmethod
    def _setup_ln2() -> float:
        """
        Returns NumPy ln2.
        """

        log2: float = np.log(2)
        return log2

    @staticmethod
    def _setup_matrix_e(size: int) -> sparse.csr_matrix:
        """
        Returns SciPy CSR sparse matrix template for matrix_e.
        """

        return sparse.csr_matrix(
            (
                np.zeros(size),
                (np.arange(size), np.arange(size)),
            )
        )

    @staticmethod
    def _setup_vector_n0(size: int) -> np.ndarray:
        """
        Returns NumPy array template for vector_n0.
        """

        return np.zeros([size], dtype=np.float64)

    def __eq__(self, other: object) -> bool:
        """
        Check whether two ``DecayMatrices`` instances are equal with ``==`` operator.
        """

        if not isinstance(other, DecayMatrices):
            return NotImplemented
        return (
            (self.atomic_masses == other.atomic_masses).all()
            and (self.decay_consts == other.decay_consts).all()
            and self.ln2 == other.ln2
            and _csr_matrix_equal(self.matrix_c, other.matrix_c)
            and _csr_matrix_equal(self.matrix_c_inv, other.matrix_c_inv)
            and _csr_matrix_equal(self.matrix_e, other.matrix_e)
            and (self.vector_n0 == other.vector_n0).all()
        )

    def __ne__(self, other: object) -> bool:
        """
        Check whether two ``DecayMatrices`` instances are not equal with ``!=`` operator.
        """

        if not isinstance(other, DecayMatrices):
            return NotImplemented
        return not self.__eq__(other)

    def __repr__(self) -> str:

        return (
            "DecayMatrices: data stored in SciPy/NumPy objects for double precision "
            "calculations."
        )


class DecayMatricesSympy(DecayMatrices):
    """
    Version of DecayMatrices using SymPy arbitrary precision operations.

    Parameters
    ----------
    atomic_masses : sympy.matrices.dense.MutableDenseMatrix
        Column vector of the atomic masses (in g/mol).
    decay_consts : sympy.matrices.dense.MutableDenseMatrix
        Column vector of the decay constants (in s\\ :sup:`-1`).
    matrix_c : sympy.matrices.sparse.MutableSparseMatrix
        A pre-calculated sparse lower traingular matrix used in decay calculations.
    matrix_c_inv : sympy.matrices.sparse.MutableSparseMatrix
        The inverse of matrix_c, also used in decay calculations.

    Attributes
    ----------
    atomic_masses : sympy.matrices.dense.MutableDenseMatrix
        Column vector of the atomic masses (in g/mol).
    decay_consts : sympy.matrices.dense.MutableDenseMatrix
        Column vector of the decay constants (in s\\ :sup:`-1`).
    ln2: log
        Constant natural logarithm of 2.
    matrix_c : sympy.matrices.sparse.MutableSparseMatrix
        A precalculated sparse lower triangular matrix used in decay calculations.
    matrix_c_inv : sympy.matrices.sparse.MutableSparseMatrix
        The inverse of matrix_c, also used in decay calculations.
    matrix_e : sympy.matrices.sparse.MutableSparseMatrix
        The matrix exponential that is used in radioactive decay calculations. It is a diagonal
        matrix that is pre-allocted for performance reasons.
    vector_n0 : sympy.matrices.dense.MutableDenseMatrix
        Column vector for the number of atoms of each nuclide. It is pre-allocted here to improve
        the performance of decay calculations.

    """

    @staticmethod
    def _setup_ln2() -> log:
        """
        Returns SymPy version of ln2.
        """

        return log(2)

    @staticmethod
    def _setup_matrix_e(size: int) -> Matrix:
        """
        Returns template for SymPy version of matrix_e.
        """

        return SparseMatrix.zeros(size, size)

    @staticmethod
    def _setup_vector_n0(size: int) -> Matrix:
        """
        Returns template for SymPy version of vector_n0.
        """

        return Matrix.zeros(size, 1)

    def __eq__(self, other: object) -> bool:
        """
        Check whether two ``DecayMatricesSympy`` instances are equal with ``==`` operator.
        """

        if not isinstance(other, DecayMatricesSympy):
            return NotImplemented
        return (
            self.atomic_masses == other.atomic_masses
            and self.decay_consts == other.decay_consts
            and self.ln2 == other.ln2
            and self.matrix_c == other.matrix_c
            and self.matrix_c_inv == other.matrix_c_inv
            and self.matrix_e == other.matrix_e
            and self.vector_n0 == other.vector_n0
        )

    def __repr__(self) -> str:

        return (
            "DecayMatricesSympy: data stored in SymPy objects for arbitrary-precision "
            "calculations."
        )


class DecayData:
    """
    Instances of DecayData store a complete radioactive decay dataset.

    Parameters
    ----------
    dataset_name : str
        Name of the decay dataset.
    dir_path : str or None, optional
        Path to the directory containing the decay dataset files. Use None if the data are bundled
        as a sub-package of ``radioactivedecay`` (default is None).
    load_sympy : bool, optional
        Load SymPy version of the decay data for arbitrary-precision decay calculations (default is
        False).

    Attributes
    ----------
    dataset_name : str
        Name of the decay dataset.
    float_quantity_converter : QuantityConverter
        Convert between quantities using floating point operations.
    float_unit_converter : UnitConverterFloat
        Convert within units using floating point operations.
    hldata : numpy.ndarray
        List of tuples containing half-life floats, time unit strings and readable format half-life
        strings.
    nuclides : numpy.ndarray
        NumPy array of nuclides in the dataset (string format is 'H-3', etc.).
    nuclide_dict : dict
        Dictionary containing nuclide strings as keys and positions in the matrices as values.
    prog_bfs_modes : numpy.ndarray
        NumPy array of dictionaries with direct progeny as keys and lists with branching fraction
        and decay mode data as values.
    scipy_data : DecayMatrices
        Dataset of double precision decay matrices (SciPy/NumPy objects).
    sympy_data : None or DecayMatricesSympy
        Dataset of arbitrary-precision decay matrices (SymPy objects). Or None if this functionality
        is not used.
    sympy_quantity_converter : None or QuantityConverterSympy
        Convert between quantities using SymPy arbitrary precision operations.
    sympy_unit_converter : None or UnitConverterSympy
        Convert within units using SymPy arbitrary precision operations.

    """

    def __init__(
        self,
        dataset_name: str,
        dir_path: Optional[str] = None,
        load_sympy: bool = False,
    ) -> None:

        self.dataset_name = dataset_name

        if dir_path is None:
            data = np.load(
                _get_package_filepath(self.dataset_name, "decay_data.npz"),
                allow_pickle=True,
            )
        else:
            data = np.load(dir_path + "/decay_data.npz", allow_pickle=True)

        self.float_unit_converter = UnitConverterFloat(data["year_conv"])
        self.hldata = data["hldata"]
        self.nuclides = data["radionuclides"]
        self.nuclide_dict = dict(zip(self.nuclides, list(range(0, self.nuclides.size))))
        self.prog_bfs_modes = data["prog_bfs_modes"]

        decay_consts = np.array(
            [
                np.log(2)
                / self.float_unit_converter.time_unit_conv(
                    hl[0], units_from=hl[1], units_to="s"
                )
                for hl in self.hldata
            ]
        )

        if dir_path is None:
            matrix_c = sparse.load_npz(
                _get_package_filepath(self.dataset_name, "c_scipy.npz")
            )
            matrix_c_inv = sparse.load_npz(
                _get_package_filepath(self.dataset_name, "c_inv_scipy.npz")
            )
        else:
            matrix_c = sparse.load_npz(dir_path + "/c_scipy.npz")
            matrix_c_inv = sparse.load_npz(dir_path + "/c_inv_scipy.npz")
        self.scipy_data = DecayMatrices(
            data["masses"], decay_consts, matrix_c, matrix_c_inv
        )

        self.float_quantity_converter = QuantityConverter(
            self.nuclide_dict,
            self.scipy_data.atomic_masses,
            self.scipy_data.decay_consts,
        )

        if load_sympy:
            if dir_path is None:
                atomic_masses = _load_package_pickle_file(
                    self.dataset_name, "atomic_masses_sympy.pickle"
                )
                decay_consts = _load_package_pickle_file(
                    self.dataset_name, "decay_consts_sympy.pickle"
                )
                matrix_c = _load_package_pickle_file(
                    self.dataset_name, "c_sympy.pickle"
                )
                matrix_c_inv = _load_package_pickle_file(
                    self.dataset_name, "c_inv_sympy.pickle"
                )
                year_conv = _load_package_pickle_file(
                    self.dataset_name, "year_conversion_sympy.pickle"
                )
            else:
                with open(dir_path + "/atomic_masses_sympy.pickle", "rb") as file:
                    atomic_masses = pickle.load(file)
                with open(dir_path + "/decay_consts_sympy.pickle", "rb") as file:
                    decay_consts = pickle.load(file)
                with open(dir_path + "/c_sympy.pickle", "rb") as file:
                    matrix_c = pickle.load(file)
                with open(dir_path + "/c_inv_sympy.pickle", "rb") as file:
                    matrix_c_inv = pickle.load(file)
                with open(dir_path + "/year_conversion_sympy.pickle", "rb") as file:
                    year_conv = pickle.load(file)

            self.sympy_data: Optional[DecayMatricesSympy] = DecayMatricesSympy(
                atomic_masses,
                decay_consts,
                matrix_c,
                matrix_c_inv,
            )
            self.sympy_unit_converter: Optional[
                UnitConverterSympy
            ] = UnitConverterSympy(year_conv)
            self.sympy_quantity_converter: Optional[
                QuantityConverterSympy
            ] = QuantityConverterSympy(
                self.nuclide_dict,
                self.sympy_data.atomic_masses,
                self.sympy_data.decay_consts,
            )
        else:
            self.sympy_data = None
            self.sympy_unit_converter = None
            self.sympy_quantity_converter = None

    def half_life(self, nuclide: str, units: str = "s") -> Union[float, str]:
        """
        Returns the half-life of the nuclide as a float in your chosen units, or as
        a human-readable string with appropriate units.

        Parameters
        ----------
        nuclide : str
            Nuclide string.
        units : str, optional
            Units for half-life. Options are 'ps', 'ns', 'μs', 'us', 'ms', 's', 'm', 'h', 'd', 'y',
            'ky', 'My', 'By', 'Gy', 'Ty', 'Py', and common spelling variations. Default is 's', i.e.
            seconds. Use 'readable' to get a string of the half-life in human-readable units.

        Returns
        -------
        float or str
            Nuclide half-life.

        Examples
        --------
        >>> rd.DEFAULTDATA.half_life('Rn-222', 'd')
        3.8235
        >>> rd.DEFAULTDATA.half_life('H-3')
        388781329.30560005
        >>> rd.DEFAULTDATA.half_life('H-3', 'readable')
        '12.32 y'
        >>> rd.DEFAULTDATA.half_life('He-3')
        inf
        >>> rd.DEFAULTDATA.half_life('He-3', 'readable')
        'stable'

        """

        nuclide = parse_nuclide(nuclide, self.nuclides, self.dataset_name)
        half_life: float
        unit: str
        readable_str: str
        half_life, unit, readable_str = self.hldata[self.nuclide_dict[nuclide]]

        if units == "readable":
            return readable_str

        return (
            half_life
            if unit == units
            else self.float_unit_converter.time_unit_conv(
                half_life,
                units_from=unit,
                units_to=units,
            )
        )

    def branching_fraction(self, parent: str, progeny: str) -> float:
        """
        Returns the branching fraction for parent to progeny (if it exists).

        Parameters
        ----------
        parent : str
            Nuclide string of the parent.
        progeny : str
            Nuclide string of the progeny (can be stable or radioactive nuclide).

        Returns
        -------
        float
            Branching fraction (or zero if progeny is not actually a direct progeny of parent).

        Examples
        --------
        >>> rd.DEFAULTDATA.branching_fraction('K-40', 'Ca-40')
        0.8914

        """

        parent = parse_nuclide(parent, self.nuclides, self.dataset_name)
        progeny = parse_nuclide(progeny, self.nuclides, self.dataset_name)
        if progeny in self.prog_bfs_modes[self.nuclide_dict[parent]]:
            branching_fraction: float = self.prog_bfs_modes[self.nuclide_dict[parent]][
                progeny
            ][0]
            return branching_fraction

        return 0.0

    def decay_mode(self, parent: str, progeny: str) -> str:
        """
        Returns the type of decay mode between parent and progeny (if one exists). Note: the decay
        mode string returned is not a list of all the different radiation types emitted during the
        decay process between parent and progeny. It is the label defined in the decay dataset to
        classify the decay type (e.g. '\u03b1', '\u03b2-' or 'IT').

        Parameters
        ----------
        parent : str
            Nuclide string of the parent.
        progeny : str
            Nuclide string of the progeny.

        Returns
        -------
        str
            Decay mode (or '' if progeny is not actually a direct progeny of the parent).

        Examples
        --------
        >>> rd.DEFAULTDATA.decay_mode('K-40', 'Ca-40')
        '\u03b2-'

        """

        parent = parse_nuclide(parent, self.nuclides, self.dataset_name)
        progeny = parse_nuclide(progeny, self.nuclides, self.dataset_name)
        if progeny in self.prog_bfs_modes[self.nuclide_dict[parent]]:
            decay_mode: str = self.prog_bfs_modes[self.nuclide_dict[parent]][progeny][1]
            return decay_mode

        return ""

    def __eq__(self, other: object) -> bool:
        """
        Check whether two ``DecayData`` instances are equal with ``==`` operator.
        """

        if not isinstance(other, DecayData):
            return NotImplemented
        return (
            self.dataset_name == other.dataset_name
            and self.float_unit_converter == other.float_unit_converter
            and (self.hldata == other.hldata).all()
            and (self.nuclides == other.nuclides).all()
            and self.nuclide_dict == other.nuclide_dict
            and (self.prog_bfs_modes == other.prog_bfs_modes).all()
            and self.scipy_data == other.scipy_data
            and self.sympy_data == other.sympy_data
            and self.sympy_unit_converter == other.sympy_unit_converter
        )

    def __ne__(self, other: object) -> bool:
        """
        Check whether two ``DecayData`` instances are not equal with ``!=`` operator.
        """

        if not isinstance(other, DecayData):
            return NotImplemented
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return (
            "Decay dataset: " + self.dataset_name + ", contains SymPy data: False"
            if self.sympy_data is None
            else "Decay dataset: " + self.dataset_name + ", contains SymPy data: True"
        )


DEFAULTDATA = DecayData("icrp107_ame2020_nubase2020", load_sympy=True)



"""
Unit tests for decaydata.py functions, classes and methods.
"""

import unittest
import numpy as np
from scipy import sparse
from sympy import Integer, log, Matrix
from sympy.matrices import SparseMatrix
from radioactivedecay import decaydata, icrp107_ame2020_nubase2020


class TestDecayDataFunctions(unittest.TestCase):
    """
    Unit tests for the decaydata.py functions.
    """

    # pylint: disable=protected-access

    def test__csr_matrix_equal(self) -> None:
        """
        Test function to check equality of two SciPy Compressed Sparse Row (CSR) matrices.
        """

        matrix_a = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_b = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c = sparse.csr_matrix(([1.0], ([1], [0])), shape=(2, 2))
        self.assertEqual(decaydata._csr_matrix_equal(matrix_a, matrix_b), True)
        self.assertEqual(decaydata._csr_matrix_equal(matrix_a, matrix_c), False)


class TestDecayMatrices(unittest.TestCase):
    """
    Unit tests for the decaydata.py DecayMatrices class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of DecayMatrices objects.
        """

        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([1], [1])), shape=(2, 2))
        decay_mats = decaydata.DecayMatrices(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(decay_mats.atomic_masses[0], 0.0)
        self.assertEqual(decay_mats.atomic_masses[1], 0.0)
        self.assertEqual(decay_mats.decay_consts[0], 0.0)
        self.assertEqual(decay_mats.decay_consts[1], 0.0)
        self.assertEqual(decay_mats.ln2, np.log(2))
        self.assertEqual(decay_mats.matrix_c[0, 0], 1.0)
        self.assertEqual(decay_mats.matrix_c_inv[1, 1], 1.0)
        self.assertEqual(decay_mats.matrix_e[0, 0], 0.0)
        self.assertEqual(decay_mats.matrix_e[1, 1], 0.0)
        self.assertEqual(decay_mats.vector_n0[0], 0.0)
        self.assertEqual(decay_mats.vector_n0[1], 0.0)

    def test_decaymatrices___eq__(self) -> None:
        """
        Test DecayMatrices equality.
        """

        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        decay_mats_a = decaydata.DecayMatrices(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        decay_mats_b = decaydata.DecayMatrices(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(decay_mats_a, decay_mats_b)

        self.assertFalse(decay_mats_a == "random object")

    def test_decaymatrices___ne__(self) -> None:
        """
        Test DecayMatrices inequal.
        """

        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        decay_mats_a = decaydata.DecayMatrices(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([2.0], ([0], [0])), shape=(2, 2))
        decay_mats_b = decaydata.DecayMatrices(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertNotEqual(decay_mats_a, decay_mats_b)

        self.assertTrue(decay_mats_a != "random object")

    def test_decaymatrices___repr__(self) -> None:
        """
        Test DecayMatrices __repr__ strings.
        """

        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([1], [1])), shape=(2, 2))
        decay_mats = decaydata.DecayMatrices(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(
            decay_mats.__repr__(),
            "DecayMatrices: data stored in SciPy/NumPy objects for double precision calculations.",
        )


class TestDecayMatricesSymPy(unittest.TestCase):
    """
    Unit tests for the decaydata.py DecayMatricesSympy class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of DecayMatricesSympy objects.
        """

        atomic_masses = Matrix.zeros(2, 1)
        decay_consts = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(3)
        decay_mats = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(decay_mats.atomic_masses[0], 0.0)
        self.assertEqual(decay_mats.atomic_masses[1], 0.0)
        self.assertEqual(decay_mats.decay_consts[0], 0.0)
        self.assertEqual(decay_mats.decay_consts[1], 0.0)
        self.assertEqual(decay_mats.ln2, log(2))
        self.assertEqual(decay_mats.matrix_c[0, 0], Integer(2))
        self.assertEqual(decay_mats.matrix_c_inv[1, 1], Integer(3))
        self.assertEqual(decay_mats.matrix_e[0, 0], Integer(0))
        self.assertEqual(decay_mats.matrix_e[1, 1], Integer(0))
        self.assertEqual(decay_mats.vector_n0[0], 0.0)
        self.assertEqual(decay_mats.vector_n0[1], 0.0)

    def test___eq__(self) -> None:
        """
        Test DecayMatricesSympy equality.
        """

        atomic_masses = Matrix.zeros(2, 1)
        decay_consts = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(3)
        decay_mats_a = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        atomic_masses = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(3)
        decay_mats_b = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(decay_mats_a, decay_mats_b)

        self.assertFalse(decay_mats_a == "random object")

    def test___ne__(self) -> None:
        """
        Test DecayMatricesSympy inequality.
        """

        atomic_masses = Matrix.zeros(2, 1)
        decay_consts = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(3)
        decay_mats_a = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        atomic_masses = Matrix.zeros(2, 1)
        decay_consts = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(5)
        decay_mats_b = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertNotEqual(decay_mats_a, decay_mats_b)

        self.assertTrue(decay_mats_a != "random object")

    def test___repr__(self) -> None:
        """
        Test DecayMatricesSympy __repr__ strings.
        """

        atomic_masses = Matrix.zeros(2, 1)
        decay_consts = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(3)
        decay_mats = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(
            decay_mats.__repr__(),
            "DecayMatricesSympy: data stored in SymPy objects for arbitrary-precision "
            + "calculations.",
        )


class TestDecayData(unittest.TestCase):
    """
    Unit tests for the decaydata.py DecayData class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of DecayData objects.
        """

        # pylint: disable=too-many-statements

        # check instantiation from sub-package
        data = decaydata.DecayData("icrp107_ame2020_nubase2020", load_sympy=False)
        self.assertEqual(data.dataset_name, "icrp107_ame2020_nubase2020")
        self.assertEqual(data.hldata[0][0], 100.5)
        self.assertEqual(data.hldata[0][1], "d")
        self.assertEqual(data.hldata[-1][0], np.inf)
        self.assertEqual(data.hldata[-1][1], "s")
        self.assertEqual(data.nuclides[0], "Fm-257")
        self.assertEqual(data.nuclides[-1], "He-3")
        self.assertEqual(data.nuclide_dict["Fm-257"], 0)
        self.assertEqual(data.nuclide_dict["He-3"], 1497)
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[0], "Cf-253")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][0], 0.9979)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][1], "\u03b1")
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][0], 0.0021)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[-1].keys()), [])
        self.assertEqual(list(data.prog_bfs_modes[-1].values()), [])
        self.assertEqual(
            data.scipy_data.decay_consts[0], np.log(2) / (100.5 * 24 * 60 * 60)
        )
        self.assertEqual(data.scipy_data.ln2, np.log(2))
        self.assertEqual(data.scipy_data.matrix_c[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_c_inv[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_e[0, 0], 0.0)
        self.assertEqual(data.scipy_data.vector_n0[0], 0.0)
        self.assertEqual(data.sympy_data, None)

        # check instantiation with supplied dataset path
        data = decaydata.DecayData(
            "icrp107_ame2020_nubase2020_2",
            icrp107_ame2020_nubase2020.__path__[0],
            load_sympy=False,
        )
        self.assertEqual(data.dataset_name, "icrp107_ame2020_nubase2020_2")
        self.assertEqual(data.hldata[0][0], 100.5)
        self.assertEqual(data.hldata[0][1], "d")
        self.assertEqual(data.hldata[-1][0], np.inf)
        self.assertEqual(data.hldata[-1][1], "s")
        self.assertEqual(data.nuclides[0], "Fm-257")
        self.assertEqual(data.nuclides[-1], "He-3")
        self.assertEqual(data.nuclide_dict["Fm-257"], 0)
        self.assertEqual(data.nuclide_dict["He-3"], 1497)
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[0], "Cf-253")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][0], 0.9979)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][1], "\u03b1")
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][0], 0.0021)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[-1].keys()), [])
        self.assertEqual(list(data.prog_bfs_modes[-1].values()), [])
        self.assertEqual(
            data.scipy_data.decay_consts[0], np.log(2) / (100.5 * 24 * 60 * 60)
        )
        self.assertEqual(data.scipy_data.ln2, np.log(2))
        self.assertEqual(data.scipy_data.matrix_c[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_c_inv[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_e[0, 0], 0.0)
        self.assertEqual(data.scipy_data.vector_n0[0], 0.0)
        self.assertEqual(data.sympy_data, None)

        # check instantiation from sub-package with SymPy data
        data = decaydata.DecayData("icrp107_ame2020_nubase2020", load_sympy=True)
        self.assertEqual(data.dataset_name, "icrp107_ame2020_nubase2020")
        self.assertEqual(data.hldata[0][0], 100.5)
        self.assertEqual(data.hldata[0][1], "d")
        self.assertEqual(data.hldata[-1][0], np.inf)
        self.assertEqual(data.hldata[-1][1], "s")
        self.assertEqual(data.nuclides[0], "Fm-257")
        self.assertEqual(data.nuclides[-1], "He-3")
        self.assertEqual(data.nuclide_dict["Fm-257"], 0)
        self.assertEqual(data.nuclide_dict["He-3"], 1497)
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[0], "Cf-253")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][0], 0.9979)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][1], "\u03b1")
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][0], 0.0021)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[-1].keys()), [])
        self.assertEqual(list(data.prog_bfs_modes[-1].values()), [])
        self.assertEqual(
            data.scipy_data.decay_consts[0], np.log(2) / (100.5 * 24 * 60 * 60)
        )
        self.assertEqual(data.scipy_data.ln2, np.log(2))
        self.assertEqual(data.scipy_data.matrix_c[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_c_inv[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_e[0, 0], 0.0)
        self.assertEqual(data.scipy_data.vector_n0[0], 0.0)
        self.assertEqual(
            data.sympy_data.decay_consts[0],
            log(2) * Integer(10) / (Integer(1005) * Integer(24) * Integer(3600)),
        )
        self.assertEqual(data.sympy_data.ln2, log(2))
        self.assertEqual(data.sympy_data.matrix_c[0, 0], Integer(1))
        self.assertEqual(data.sympy_data.matrix_c_inv[0, 0], Integer(1))
        self.assertEqual(data.sympy_data.matrix_e[0, 0], Integer(0))
        self.assertEqual(data.sympy_data.vector_n0[0], Integer(0))

        # check instantiation with supplied dataset path with SymPy data
        data = decaydata.DecayData(
            "icrp107_ame2020_nubase2020_2",
            icrp107_ame2020_nubase2020.__path__[0],
            load_sympy=True,
        )
        self.assertEqual(data.dataset_name, "icrp107_ame2020_nubase2020_2")
        self.assertEqual(data.hldata[0][0], 100.5)
        self.assertEqual(data.hldata[0][1], "d")
        self.assertEqual(data.hldata[-1][0], np.inf)
        self.assertEqual(data.hldata[-1][1], "s")
        self.assertEqual(data.nuclides[0], "Fm-257")
        self.assertEqual(data.nuclides[-1], "He-3")
        self.assertEqual(data.nuclide_dict["Fm-257"], 0)
        self.assertEqual(data.nuclide_dict["He-3"], 1497)
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[0], "Cf-253")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][0], 0.9979)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][1], "\u03b1")
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][0], 0.0021)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[-1].keys()), [])
        self.assertEqual(list(data.prog_bfs_modes[-1].values()), [])
        self.assertEqual(
            data.scipy_data.decay_consts[0], np.log(2) / (100.5 * 24 * 60 * 60)
        )
        self.assertEqual(data.scipy_data.ln2, np.log(2))
        self.assertEqual(data.scipy_data.matrix_c[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_c_inv[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_e[0, 0], 0.0)
        self.assertEqual(data.scipy_data.vector_n0[0], 0.0)
        self.assertEqual(
            data.sympy_data.decay_consts[0],
            log(2) * Integer(10) / (Integer(1005) * Integer(24) * Integer(3600)),
        )
        self.assertEqual(data.sympy_data.ln2, log(2))
        self.assertEqual(data.sympy_data.matrix_c[0, 0], Integer(1))
        self.assertEqual(data.sympy_data.matrix_c_inv[0, 0], Integer(1))
        self.assertEqual(data.sympy_data.matrix_e[0, 0], Integer(0))
        self.assertEqual(data.sympy_data.vector_n0[0], Integer(0))

    def test_half_life(self) -> None:
        """
        Test DecayData half_life() method.
        """

        data = decaydata.DecayData("icrp107_ame2020_nubase2020")
        self.assertEqual(data.half_life("H-3"), 388781329.30560005)
        self.assertEqual(data.half_life("H-3", "y"), 12.32)
        self.assertEqual(data.half_life("Fm-257", "h"), 2412.0)
        self.assertEqual(data.half_life("Rn-222", "d"), 3.8235)

        self.assertEqual(data.half_life("H-3", "readable"), "12.32 y")
        self.assertEqual(data.half_life("Po-213", "readable"), "4.2 μs")
        self.assertEqual(data.half_life("Ra-219", "readable"), "10 ms")
        self.assertEqual(data.half_life("Rn-215", "readable"), "2.30 μs")
        self.assertEqual(data.half_life("U-238", "readable"), "4.468 By")

    def test_branching_fraction(self) -> None:
        """
        Test DecayData branching_fraction() method.
        """

        data = decaydata.DecayData("icrp107_ame2020_nubase2020")
        self.assertEqual(data.branching_fraction("K-40", "Ca-40"), 0.8914)
        self.assertEqual(data.branching_fraction("K-40", "H-3"), 0.0)

    def test_decay_mode(self) -> None:
        """
        Test DecayData decay_mode() method.
        """

        data = decaydata.DecayData("icrp107_ame2020_nubase2020")
        self.assertEqual(data.decay_mode("K-40", "Ca-40"), "\u03b2-")
        self.assertEqual(data.decay_mode("K-40", "H-3"), "")

    def test_decaydata___eq__(self) -> None:
        """
        Test DecayData equality.
        """

        data1 = decaydata.DecayData("icrp107_ame2020_nubase2020")
        data2 = decaydata.DecayData("icrp107_ame2020_nubase2020")
        self.assertEqual(data1, data2)

        self.assertFalse(data1 == "random object")

    def test_decaydata___ne__(self) -> None:
        """
        Test DecayData inequality.
        """

        data1 = decaydata.DecayData("icrp107_ame2020_nubase2020")
        data2 = decaydata.DecayData("icrp107_ame2020_nubase2020")
        data2.dataset_name = "icrp07"
        self.assertNotEqual(data1, data2)

        self.assertTrue(data1 != "random object")

    def test_decaydata___repr__(self) -> None:
        """
        Test DecayData __repr__ strings.
        """

        data = decaydata.DecayData("icrp107_ame2020_nubase2020")
        self.assertEqual(
            data.__repr__(),
            "Decay dataset: icrp107_ame2020_nubase2020, contains SymPy data: False",
        )

        data = decaydata.DecayData("icrp107_ame2020_nubase2020", load_sympy=True)
        self.assertEqual(
            data.__repr__(),
            "Decay dataset: icrp107_ame2020_nubase2020, contains SymPy data: True",
        )


if __name__ == "__main__":
    unittest.main()



"""
Unit tests for converts.py classes and methods.
"""

import unittest
import numpy as np
from sympy import Integer, log
from radioactivedecay.converters import (
    UnitConverterFloat,
    UnitConverterSympy,
    QuantityConverter,
    QuantityConverterSympy,
)


class TestUnitConverterFloat(unittest.TestCase):
    """
    Unit tests for the utils.py UnitConverterFloat class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of UnitConverterFloat objects.
        """

        year_conv = 365.2422
        uconv = UnitConverterFloat(year_conv)
        self.assertEqual(uconv.time_units["s"], 1.0)
        self.assertEqual(uconv.time_units["y"], 86400.0 * year_conv)
        self.assertEqual(uconv.activity_units["Bq"], 1.0)
        self.assertEqual(uconv.mass_units["g"], 1.0)
        self.assertEqual(uconv.moles_units["mol"], 1.0)

    def test_time_unit_conv_seconds(self) -> None:
        """
        Test conversion between seconds and different time units.
        """

        year_conv = 365.2422
        uconv = UnitConverterFloat(year_conv)

        self.assertEqual(uconv.time_unit_conv(1.0, "s", "s"), 1.0e0)
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "s", "ps"), 1.0e12, places=(15 - 12)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "s", "ns"), 1.0e9, places=(15 - 9)
        )
        self.assertEqual(uconv.time_unit_conv(1.0, "s", "μs"), 1.0e6)
        self.assertEqual(uconv.time_unit_conv(1.0, "s", "us"), 1.0e6)
        self.assertEqual(uconv.time_unit_conv(1.0, "s", "ms"), 1.0e3)
        self.assertEqual(uconv.time_unit_conv(1.0, "s", "m"), 1.0 / 60.0)
        self.assertEqual(uconv.time_unit_conv(1.0, "s", "h"), 1.0 / (60.0 ** 2))
        self.assertEqual(uconv.time_unit_conv(1.0, "s", "d"), 1.0 / (60.0 ** 2 * 24.0))
        self.assertEqual(
            uconv.time_unit_conv(1.0, "s", "y"),
            1.0 / (60.0 ** 2 * 24.0 * year_conv),
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "ps", "s"), 1.0e-12, places=(12 + 15)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "ns", "s"), 1.0e-9, places=(9 + 15)
        )
        self.assertEqual(uconv.time_unit_conv(1.0, "μs", "s"), 1.0e-6)
        self.assertEqual(uconv.time_unit_conv(1.0, "us", "s"), 1.0e-6)
        self.assertEqual(uconv.time_unit_conv(1.0, "ms", "s"), 1.0e-3)
        self.assertEqual(uconv.time_unit_conv(1.0, "m", "s"), 60.0)
        self.assertEqual(uconv.time_unit_conv(1.0, "h", "s"), (60.0 ** 2))
        self.assertEqual(uconv.time_unit_conv(1.0, "d", "s"), (60.0 ** 2 * 24.0))
        self.assertEqual(
            uconv.time_unit_conv(1.0, "y", "s"), (60.0 ** 2 * 24.0 * year_conv)
        )

        # Catch some incorrect time units
        with self.assertRaises(ValueError):
            uconv.time_unit_conv(1.0, "ty", "y")
        with self.assertRaises(ValueError):
            uconv.time_unit_conv(1.0, "y", "ty")
        with self.assertRaises(ValueError):
            uconv.time_unit_conv(1.0, "ty", 1.0)

    def test_time_unit_conv_spelling_variations(self) -> None:
        """
        Test spelling variations of different time units.
        """

        year_conv = 365.2422
        uconv = UnitConverterFloat(year_conv)

        self.assertEqual(uconv.time_unit_conv(1.0, "s", "sec"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "s", "second"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "s", "seconds"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "h", "hr"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "h", "hour"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "h", "hours"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "d", "day"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "d", "days"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "y", "yr"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "y", "year"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "y", "years"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "sec", "s"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "second", "s"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "seconds", "s"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "hr", "h"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "hour", "h"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "hours", "h"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "day", "d"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "days", "d"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "yr", "y"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "year", "y"), 1.0e0)
        self.assertEqual(uconv.time_unit_conv(1.0, "years", "y"), 1.0e0)

    def test_time_unit_conv_year_prefixes(self) -> None:
        """
        Test conversions between different year prefixes.
        """

        year_conv = 365.2422
        uconv = UnitConverterFloat(year_conv)

        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "y", "ky"), 1.0e-3, places=(3 + 15)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "y", "My"), 1.0e-6, places=(6 + 15)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "y", "By"), 1.0e-9, places=(9 + 15)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "y", "Gy"), 1.0e-9, places=(9 + 15)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "y", "Ty"), 1.0e-12, places=(12 + 15)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "y", "Py"), 1.0e-15, places=(15 + 15)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "ky", "y"), 1.0e3, places=(15 - 3)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "My", "y"), 1.0e6, places=(15 - 6)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "Gy", "y"), 1.0e9, places=(15 - 9)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "Ty", "y"), 1.0e12, places=(15 - 12)
        )
        self.assertAlmostEqual(
            uconv.time_unit_conv(1.0, "Py", "y"), 1.0e15, places=(15 - 15)
        )

    def test_activity_unit_conv(self) -> None:
        """
        Test conversions between activity units.
        """

        uconv = UnitConverterFloat(365.2422)

        self.assertEqual(uconv.activity_unit_conv(1.0, "Bq", "Bq"), 1.0e0)
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "pBq"), 1.0e12, places=(15 - 12)
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "nBq"), 1.0e9, places=(15 - 9)
        )
        self.assertEqual(uconv.activity_unit_conv(1.0, "Bq", "μBq"), 1.0e6)
        self.assertEqual(uconv.activity_unit_conv(1.0, "Bq", "uBq"), 1.0e6)
        self.assertEqual(uconv.activity_unit_conv(1.0, "Bq", "mBq"), 1.0e3)
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "kBq"), 1.0e-3, places=(3 + 15)
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "MBq"), 1.0e-6, places=(6 + 15)
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "GBq"), 1.0e-9, places=(9 + 15)
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "TBq"), 1.0e-12, places=(12 + 15)
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "PBq"), 1.0e-15, places=(15 + 15)
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "EBq"), 1.0e-18, places=(18 + 15)
        )

        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "pCi"),
            1.0e12 / 3.7e10,
            places=(15 - 12),
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "nCi"), 1.0e9 / 3.7e10, places=(15 - 9)
        )
        self.assertEqual(uconv.activity_unit_conv(1.0, "Bq", "μCi"), 1.0e6 / 3.7e10)
        self.assertEqual(uconv.activity_unit_conv(1.0, "Bq", "uCi"), 1.0e6 / 3.7e10)
        self.assertEqual(uconv.activity_unit_conv(1.0, "Bq", "mCi"), 1.0e3 / 3.7e10)
        self.assertEqual(uconv.activity_unit_conv(1.0, "Bq", "Ci"), 1.0 / 3.7e10)
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "kCi"), 1.0e-3 / 3.7e10, places=(3 + 15)
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "MCi"), 1.0e-6 / 3.7e10, places=(6 + 15)
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "GCi"), 1.0e-9 / 3.7e10, places=(9 + 15)
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "TCi"),
            1.0e-12 / 3.7e10,
            places=(12 + 15),
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "PCi"),
            1.0e-15 / 3.7e10,
            places=(15 + 15),
        )
        self.assertAlmostEqual(
            uconv.activity_unit_conv(1.0, "Bq", "ECi"),
            1.0e-18 / 3.7e10,
            places=(18 + 15),
        )

        self.assertEqual(uconv.activity_unit_conv(1.0, "Bq", "dpm"), 1.0 / 60.0)

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            uconv.activity_unit_conv(1.0, "tBq", "Bq")
        with self.assertRaises(ValueError):
            uconv.activity_unit_conv(1.0, "Bq", "tBq")
        with self.assertRaises(ValueError):
            uconv.activity_unit_conv(1.0, "tBq", 1.0)

    def test_mass_unit_conv(self) -> None:
        """
        Test conversions between mass units.
        """

        uconv = UnitConverterFloat(365.2422)

        self.assertEqual(uconv.mass_unit_conv(1.0, "g", "g"), 1.0e0)
        self.assertAlmostEqual(
            uconv.mass_unit_conv(1.0, "g", "pg"), 1.0e12, places=(15 - 12)
        )
        self.assertAlmostEqual(
            uconv.mass_unit_conv(1.0, "g", "ng"), 1.0e9, places=(15 - 9)
        )
        self.assertEqual(uconv.mass_unit_conv(1.0, "g", "μg"), 1.0e6)
        self.assertEqual(uconv.mass_unit_conv(1.0, "g", "ug"), 1.0e6)
        self.assertEqual(uconv.mass_unit_conv(1.0, "g", "mg"), 1.0e3)
        self.assertAlmostEqual(
            uconv.mass_unit_conv(1.0, "g", "kg"), 1.0e-3, places=(3 + 15)
        )
        self.assertAlmostEqual(
            uconv.mass_unit_conv(1.0, "g", "Mg"), 1.0e-6, places=(6 + 15)
        )
        self.assertAlmostEqual(
            uconv.mass_unit_conv(1.0, "g", "t"), 1.0e-6, places=(6 + 15)
        )
        self.assertAlmostEqual(
            uconv.mass_unit_conv(1.0, "g", "ton"), 1.0e-6, places=(6 + 15)
        )

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            uconv.mass_unit_conv(1.0, "tg", "g")
        with self.assertRaises(ValueError):
            uconv.mass_unit_conv(1.0, "g", "tg")
        with self.assertRaises(ValueError):
            uconv.mass_unit_conv(1.0, "tg", 1.0)

    def test_moles_unit_conv(self) -> None:
        """
        Test conversions between moles orders of magnitude.
        """

        uconv = UnitConverterFloat(365.2422)

        self.assertEqual(uconv.moles_unit_conv(1.0, "mol", "mol"), 1.0e0)
        self.assertAlmostEqual(
            uconv.moles_unit_conv(1.0, "mol", "pmol"), 1.0e12, places=(15 - 12)
        )
        self.assertAlmostEqual(
            uconv.moles_unit_conv(1.0, "mol", "nmol"), 1.0e9, places=(15 - 9)
        )
        self.assertEqual(uconv.moles_unit_conv(1.0, "mol", "μmol"), 1.0e6)
        self.assertEqual(uconv.moles_unit_conv(1.0, "mol", "umol"), 1.0e6)
        self.assertEqual(uconv.moles_unit_conv(1.0, "mol", "mmol"), 1.0e3)
        self.assertAlmostEqual(
            uconv.moles_unit_conv(1.0, "mol", "kmol"), 1.0e-3, places=(3 + 15)
        )
        self.assertAlmostEqual(
            uconv.moles_unit_conv(1.0, "mol", "Mmol"), 1.0e-6, places=(6 + 15)
        )

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            uconv.moles_unit_conv(1.0, "tmol", "mol")
        with self.assertRaises(ValueError):
            uconv.moles_unit_conv(1.0, "mol", "tmol")
        with self.assertRaises(ValueError):
            uconv.moles_unit_conv(1.0, "tmol", 1.0)

    def test___eq__(self) -> None:
        """
        Test UnitConverterFloat equality.
        """

        uc1 = UnitConverterFloat(365.2422)
        uc2 = UnitConverterFloat(365.2422)
        self.assertEqual(uc1, uc2)

        self.assertFalse(uc1 == "random object")

    def test___neq__(self) -> None:
        """
        Test UnitConverterFloat inequality.
        """

        uc1 = UnitConverterFloat(365.2422)
        uc2 = UnitConverterFloat(365.25)
        self.assertNotEqual(uc1, uc2)

        self.assertTrue(uc1 != "random object")

    def test___repr__(self) -> None:
        """
        Test UnitConverterFloat __repr__ strings.
        """

        uconv = UnitConverterFloat(365.2422)
        self.assertEqual(
            uconv.__repr__(), "UnitConverterFloat using 365.2422 days in a year."
        )


class TestUnitConverterSympy(unittest.TestCase):
    """
    Unit tests for the utils.py UnitConverterSympy class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of UnitConverterSympy objects.
        """

        year_conv = Integer(3652422) / Integer(10000)
        ucs = UnitConverterSympy(year_conv)
        self.assertEqual(ucs.time_units["s"], Integer(1))
        self.assertEqual(ucs.time_units["y"], Integer(86400) * year_conv)
        self.assertEqual(ucs.activity_units["Bq"], Integer(1))
        self.assertEqual(ucs.mass_units["g"], Integer(1))
        self.assertEqual(ucs.moles_units["mol"], Integer(1))

    def test_time_unit_conv(self) -> None:
        """
        Test of the SymPy version of time_unit_conv().
        """

        year_conv = Integer(3652422) / 10000
        ucs = UnitConverterSympy(year_conv)

        self.assertEqual(ucs.time_unit_conv(Integer(1), "ps", "ns"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "ns", "us"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "μs", "ms"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "us", "ms"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "ms", "s"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "s", "m"), 1 / Integer(60))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "m", "h"), 1 / Integer(60))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "h", "d"), 1 / Integer(24))
        self.assertEqual(
            ucs.time_unit_conv(Integer(1), "d", "y"), 10000 / Integer(3652422)
        )
        self.assertEqual(ucs.time_unit_conv(Integer(1), "y", "ky"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "ky", "My"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "My", "By"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "My", "Gy"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "By", "Ty"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "Gy", "Ty"), 1 / Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "Ty", "Py"), 1 / Integer(1000))

        self.assertEqual(ucs.time_unit_conv(Integer(1), "ns", "ps"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "μs", "ns"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "us", "ns"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "ms", "us"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "s", "ms"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "m", "s"), Integer(60))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "h", "m"), Integer(60))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "d", "h"), Integer(24))
        self.assertEqual(
            ucs.time_unit_conv(Integer(1), "y", "d"), Integer(3652422) / 10000
        )
        self.assertEqual(ucs.time_unit_conv(Integer(1), "ky", "y"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "My", "ky"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "By", "My"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "Gy", "My"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "Ty", "By"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "Ty", "Gy"), Integer(1000))
        self.assertEqual(ucs.time_unit_conv(Integer(1), "Py", "Ty"), Integer(1000))

        # Catch some incorrect time units
        with self.assertRaises(ValueError):
            ucs.time_unit_conv(Integer(1), "ty", "y")
        with self.assertRaises(ValueError):
            ucs.time_unit_conv(Integer(1), "y", "ty")
        with self.assertRaises(ValueError):
            ucs.time_unit_conv(Integer(1), "ty", Integer(1))

    def test_activity_unit_conv(self) -> None:
        """
        Test of the SymPy version of activity_unit_conv().
        """

        year_conv = Integer(3652422) / 10000
        ucs = UnitConverterSympy(year_conv)

        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "pBq", "nBq"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "nBq", "μBq"), 1 / Integer(1000)
        )
        self.assertEqual(ucs.activity_unit_conv(Integer(1), "μBq", "uBq"), Integer(1))
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "uBq", "mBq"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "mBq", "Bq"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "Bq", "kBq"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "kBq", "MBq"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "MBq", "GBq"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "GBq", "TBq"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "TBq", "PBq"), 1 / Integer(1000)
        )

        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "pBq", "pCi"), 1 / Integer(37000000000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "pCi", "nCi"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "nCi", "μCi"), 1 / Integer(1000)
        )
        self.assertEqual(ucs.activity_unit_conv(Integer(1), "μCi", "uCi"), Integer(1))
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "uCi", "mCi"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "mCi", "Ci"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "Ci", "kCi"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "kCi", "MCi"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "MCi", "GCi"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "GCi", "TCi"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "TCi", "PCi"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "PCi", "ECi"), 1 / Integer(1000)
        )

        self.assertEqual(
            ucs.activity_unit_conv(Integer(1), "Bq", "dpm"), 1 / Integer(60)
        )

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            ucs.activity_unit_conv(Integer(1), "tBq", "Bq")
        with self.assertRaises(ValueError):
            ucs.activity_unit_conv(Integer(1), "Bq", "tBq")
        with self.assertRaises(ValueError):
            ucs.activity_unit_conv(Integer(1), "tBq", Integer(1))

    def test_mass_unit_conv(self) -> None:
        """
        Test of the SymPy version of mass_unit_conv().
        """

        year_conv = Integer(3652422) / 10000
        ucs = UnitConverterSympy(year_conv)

        self.assertEqual(ucs.mass_unit_conv(Integer(1), "pg", "ng"), 1 / Integer(1000))
        self.assertEqual(ucs.mass_unit_conv(Integer(1), "ng", "μg"), 1 / Integer(1000))
        self.assertEqual(ucs.mass_unit_conv(Integer(1), "μg", "ug"), Integer(1))
        self.assertEqual(ucs.mass_unit_conv(Integer(1), "ug", "mg"), 1 / Integer(1000))
        self.assertEqual(ucs.mass_unit_conv(Integer(1), "mg", "g"), 1 / Integer(1000))
        self.assertEqual(ucs.mass_unit_conv(Integer(1), "g", "kg"), 1 / Integer(1000))
        self.assertEqual(ucs.mass_unit_conv(Integer(1), "kg", "Mg"), 1 / Integer(1000))
        self.assertEqual(ucs.mass_unit_conv(Integer(1), "Mg", "t"), Integer(1))
        self.assertEqual(ucs.mass_unit_conv(Integer(1), "Mg", "ton"), Integer(1))

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            ucs.mass_unit_conv(Integer(1), "tg", "g")
        with self.assertRaises(ValueError):
            ucs.mass_unit_conv(Integer(1), "g", "tg")
        with self.assertRaises(ValueError):
            ucs.mass_unit_conv(Integer(1), "tg", Integer(1))

    def test_moles_unit_conv(self) -> None:
        """
        Test of the SymPy version of moles_unit_conv().
        """

        year_conv = Integer(3652422) / 10000
        ucs = UnitConverterSympy(year_conv)

        self.assertEqual(
            ucs.moles_unit_conv(Integer(1), "pmol", "nmol"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.moles_unit_conv(Integer(1), "nmol", "μmol"), 1 / Integer(1000)
        )
        self.assertEqual(ucs.moles_unit_conv(Integer(1), "μmol", "umol"), Integer(1))
        self.assertEqual(
            ucs.moles_unit_conv(Integer(1), "umol", "mmol"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.moles_unit_conv(Integer(1), "mmol", "mol"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.moles_unit_conv(Integer(1), "mol", "kmol"), 1 / Integer(1000)
        )
        self.assertEqual(
            ucs.moles_unit_conv(Integer(1), "kmol", "Mmol"), 1 / Integer(1000)
        )

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            ucs.moles_unit_conv(Integer(1), "tmol", "mol")
        with self.assertRaises(ValueError):
            ucs.moles_unit_conv(Integer(1), "mol", "tmol")
        with self.assertRaises(ValueError):
            ucs.moles_unit_conv(Integer(1), "tmol", Integer(1))

    def test___repr__(self) -> None:
        """
        Test UnitConverterSympy __repr__ strings.
        """

        year_conv = Integer(3652422) / 10000
        ucs = UnitConverterSympy(year_conv)
        self.assertEqual(
            ucs.__repr__(), "UnitConverterSympy using 1826211/5000 days in a year."
        )


class TestQuantityConverter(unittest.TestCase):
    """
    Unit tests for the utils.py QuantityConverter class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of QuantityConverter objects.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = np.array([3.01604928132, 3.01602932197])
        decay_consts = np.array([1.7828715741004621e-09, 0.0])
        qconv = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)
        self.assertEqual(qconv.nuclide_dict["He-3"], 1)
        self.assertEqual(qconv.atomic_masses[0], 3.01604928132)
        self.assertEqual(qconv.decay_consts[1], 0)

    def test_activity_to_number(self) -> None:
        """
        Test the conversion of activity in Bq to number of atoms.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = np.array([3.01604928132, 3.01602932197])
        decay_consts = np.array([1.7828715741004621e-09, 0.0])
        qconv = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)
        self.assertEqual(qconv.activity_to_number("H-3", 1.0), 560892895.7794082)

    def test_mass_to_number(self) -> None:
        """
        Test the conversion of mass in grams to number of atoms.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = np.array([3.01604928132, 3.01602932197])
        decay_consts = np.array([1.7828715741004621e-09, 0.0])
        qconv = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(qconv.mass_to_number("H-3", 1.0), 1.996698395247825e23)
        self.assertEqual(qconv.mass_to_number("He-3", 1.0), 1.9967116089131645e23)

    def test_moles_to_number(self) -> None:
        """
        Test the conversion of number of moles to number of atoms.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = np.array([3.01604928132, 3.01602932197])
        decay_consts = np.array([1.7828715741004621e-09, 0.0])
        qconv = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(qconv.moles_to_number(1.0), 6.02214076e23)
        self.assertEqual(qconv.moles_to_number(0.0), 0.0)

    def test_number_to_activity(self) -> None:
        """
        Test the conversion of number of atoms to activity in Bq.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = np.array([3.01604928132, 3.01602932197])
        decay_consts = np.array([1.7828715741004621e-09, 0.0])
        qconv = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(qconv.number_to_activity("H-3", 560892895.7794082), 1.0)
        self.assertEqual(qconv.number_to_activity("He-3", 1.0), 0.0)
        self.assertEqual(qconv.number_to_activity("He-3", 0.0), 0.0)

    def test_number_to_mass(self) -> None:
        """
        Test the conversion of number of atoms to mass in grams.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = np.array([3.01604928132, 3.01602932197])
        decay_consts = np.array([1.7828715741004621e-09, 0.0])
        qconv = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(qconv.number_to_mass("H-3", 1.996698395247825e23), 1.0)
        self.assertEqual(qconv.number_to_mass("He-3", 1.9967116089131645e23), 1.0)

    def test_number_to_moles(self) -> None:
        """
        Test the conversion of number of atoms to number of moles.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = np.array([3.01604928132, 3.01602932197])
        decay_consts = np.array([1.7828715741004621e-09, 0.0])
        qconv = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(qconv.number_to_moles(6.02214076e23), 1.0)
        self.assertEqual(qconv.number_to_moles(0.0), 0.0)

    def test___eq__(self) -> None:
        """
        Test QuantityConveter equality.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = np.array([3.01604928132, 3.01602932197])
        decay_consts = np.array([1.7828715741004621e-09, 0.0])
        qconv1 = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)
        qconv2 = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)
        self.assertEqual(qconv1, qconv2)

        self.assertFalse(qconv1 == "random object")

    def test___neq__(self) -> None:
        """
        Test QuantityConveter inequality.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = np.array([3.01604928132, 3.01602932197])
        decay_consts = np.array([1.7828715741004621e-09, 0.0])
        qconv1 = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)
        decay_consts = np.array([1.7828715741004621e-09, 0.1])
        qconv2 = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)
        self.assertNotEqual(qconv1, qconv2)

        self.assertTrue(qconv1 != "random object")

    def test___repr__(self) -> None:
        """
        Test QuantityConveter __repr__ strings.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = np.array([3.01604928132, 3.01602932197])
        decay_consts = np.array([1.7828715741004621e-09, 0.0])
        qconv = QuantityConverter(nuclide_dict, atomic_masses, decay_consts)
        self.assertEqual(
            qconv.__repr__(), "QuantityConverter using double-precision floats."
        )


class TestQuantityConverterSympy(unittest.TestCase):
    """
    Unit tests for the utils.py QuantityConverterSympy class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of QuantityConverter objects.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = [
            Integer("15055351900000000000000000000000000") / Integer(75401232033),
            Integer("60221407600000000000000000000000000") / Integer(301602932197),
        ]
        decay_consts = [(Integer(625) * log(2)) / Integer(242988330816), Integer(0)]

        qcs = QuantityConverterSympy(nuclide_dict, atomic_masses, decay_consts)
        self.assertEqual(qcs.nuclide_dict["He-3"], 1)
        self.assertEqual(
            qcs.atomic_masses[0],
            Integer("15055351900000000000000000000000000") / Integer(75401232033),
        )
        self.assertEqual(qcs.decay_consts[1], Integer(0))

    def test_activity_to_number(self) -> None:
        """
        Test the conversion of activity in Bq to number of atoms.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = [
            Integer("15055351900000000000000000000000000") / Integer(75401232033),
            Integer("60221407600000000000000000000000000") / Integer(301602932197),
        ]
        decay_consts = [(Integer(625) * log(2)) / Integer(242988330816), Integer(0)]
        qcs = QuantityConverterSympy(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(
            qcs.activity_to_number("H-3", Integer(1)),
            Integer(242988330816) / (Integer(625) * log(2)),
        )

    def test_mass_to_number(self) -> None:
        """
        Test the conversion of mass in grams to number of atoms.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = [
            Integer("15055351900000000000000000000000000") / Integer(75401232033),
            Integer("60221407600000000000000000000000000") / Integer(301602932197),
        ]
        decay_consts = [(Integer(625) * log(2)) / Integer(242988330816), Integer(0)]
        qcs = QuantityConverterSympy(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(
            qcs.mass_to_number("H-3", Integer(1)),
            Integer(75401232033)
            / Integer("15055351900000000000000000000000000")
            * Integer("602214076000000000000000"),
        )
        self.assertEqual(
            qcs.mass_to_number("He-3", Integer(1)),
            Integer(301602932197)
            / Integer("60221407600000000000000000000000000")
            * Integer("602214076000000000000000"),
        )

    def test_moles_to_number(self) -> None:
        """
        Test the conversion of number of moles to number of atoms.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = [
            Integer("15055351900000000000000000000000000") / Integer(75401232033),
            Integer("60221407600000000000000000000000000") / Integer(301602932197),
        ]
        decay_consts = [(Integer(625) * log(2)) / Integer(242988330816), Integer(0)]
        qcs = QuantityConverterSympy(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(
            qcs.moles_to_number(Integer(1)), Integer("602214076000000000000000")
        )
        self.assertEqual(qcs.moles_to_number(Integer(0)), Integer(0))

    def test_number_to_activity(self) -> None:
        """
        Test the conversion of number of atoms to activity in Bq.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = [
            Integer("15055351900000000000000000000000000") / Integer(75401232033),
            Integer("60221407600000000000000000000000000") / Integer(301602932197),
        ]
        decay_consts = [(Integer(625) * log(2)) / Integer(242988330816), Integer(0)]
        qcs = QuantityConverterSympy(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(
            qcs.number_to_activity("H-3", Integer(1)),
            Integer(625) * log(2) / Integer(242988330816),
        )
        self.assertEqual(qcs.number_to_activity("He-3", Integer(0)), Integer(0))
        self.assertEqual(qcs.number_to_activity("He-3", Integer(0)), Integer(0))

    def test_number_to_mass(self) -> None:
        """
        Test the conversion of number of atoms to mass in grams.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = [
            Integer("15055351900000000000000000000000000") / Integer(75401232033),
            Integer("60221407600000000000000000000000000") / Integer(301602932197),
        ]
        decay_consts = [(Integer(625) * log(2)) / Integer(242988330816), Integer(0)]
        qcs = QuantityConverterSympy(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(
            qcs.number_to_mass("H-3", Integer(1)),
            Integer("15055351900000000000000000000000000")
            / Integer(75401232033)
            / Integer("602214076000000000000000"),
        )
        self.assertEqual(
            qcs.number_to_mass("He-3", Integer(1)),
            Integer("60221407600000000000000000000000000")
            / Integer(301602932197)
            / Integer("602214076000000000000000"),
        )

    def test_number_to_moles(self) -> None:
        """
        Test the conversion of number of atoms to number of moles.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = [
            Integer("15055351900000000000000000000000000") / Integer(75401232033),
            Integer("60221407600000000000000000000000000") / Integer(301602932197),
        ]
        decay_consts = [(Integer(625) * log(2)) / Integer(242988330816), Integer(0)]
        qcs = QuantityConverterSympy(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(
            qcs.number_to_moles(Integer("602214076000000000000000")), Integer(1)
        )
        self.assertEqual(qcs.number_to_moles(Integer(0)), Integer(0))

    def test___eq__(self) -> None:
        """
        Test QuantityConveterSympy equality.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = [
            Integer("15055351900000000000000000000000000") / Integer(75401232033),
            Integer("60221407600000000000000000000000000") / Integer(301602932197),
        ]
        decay_consts = [(Integer(625) * log(2)) / Integer(242988330816), Integer(0)]
        qconv1 = QuantityConverterSympy(nuclide_dict, atomic_masses, decay_consts)
        qconv2 = QuantityConverterSympy(nuclide_dict, atomic_masses, decay_consts)
        self.assertEqual(qconv1, qconv2)

        self.assertFalse(qconv1 == "random object")

    def test___repr__(self) -> None:
        """
        Test QuantityConveterSympy __repr__ strings.
        """

        nuclide_dict = {"H-3": 0, "He-3": 1}
        atomic_masses = [
            Integer("15055351900000000000000000000000000") / Integer(75401232033),
            Integer("60221407600000000000000000000000000") / Integer(301602932197),
        ]
        decay_consts = [(Integer(625) * log(2)) / Integer(242988330816), Integer(0)]
        qcs = QuantityConverterSympy(nuclide_dict, atomic_masses, decay_consts)

        self.assertEqual(
            qcs.__repr__(),
            "QuantityConverterSympy using SymPy arbitrary precision calculations.",
        )


if __name__ == "__main__":
    unittest.main()



"""
The utils module contains functions to parse nuclide strings and manipulate lists and dictionaries.

The code examples shown in the docstrings assume the ``radioactivedecay`` package has been imported
as:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

"""

from typing import Dict, List, Union
from sympy.core.expr import Expr


Z_DICT = {
    1: "H",
    2: "He",
    3: "Li",
    4: "Be",
    5: "B",
    6: "C",
    7: "N",
    8: "O",
    9: "F",
    10: "Ne",
    11: "Na",
    12: "Mg",
    13: "Al",
    14: "Si",
    15: "P",
    16: "S",
    17: "Cl",
    18: "Ar",
    19: "K",
    20: "Ca",
    21: "Sc",
    22: "Ti",
    23: "V",
    24: "Cr",
    25: "Mn",
    26: "Fe",
    27: "Co",
    28: "Ni",
    29: "Cu",
    30: "Zn",
    31: "Ga",
    32: "Ge",
    33: "As",
    34: "Se",
    35: "Br",
    36: "Kr",
    37: "Rb",
    38: "Sr",
    39: "Y",
    40: "Zr",
    41: "Nb",
    42: "Mo",
    43: "Tc",
    44: "Ru",
    45: "Rh",
    46: "Pd",
    47: "Ag",
    48: "Cd",
    49: "In",
    50: "Sn",
    51: "Sb",
    52: "Te",
    53: "I",
    54: "Xe",
    55: "Cs",
    56: "Ba",
    57: "La",
    58: "Ce",
    59: "Pr",
    60: "Nd",
    61: "Pm",
    62: "Sm",
    63: "Eu",
    64: "Gd",
    65: "Tb",
    66: "Dy",
    67: "Ho",
    68: "Er",
    69: "Tm",
    70: "Yb",
    71: "Lu",
    72: "Hf",
    73: "Ta",
    74: "W",
    75: "Re",
    76: "Os",
    77: "Ir",
    78: "Pt",
    79: "Au",
    80: "Hg",
    81: "Tl",
    82: "Pb",
    83: "Bi",
    84: "Po",
    85: "At",
    86: "Rn",
    87: "Fr",
    88: "Ra",
    89: "Ac",
    90: "Th",
    91: "Pa",
    92: "U",
    93: "Np",
    94: "Pu",
    95: "Am",
    96: "Cm",
    97: "Bk",
    98: "Cf",
    99: "Es",
    100: "Fm",
    101: "Md",
    102: "No",
    103: "Lr",
    104: "Rf",
    105: "Db",
    106: "Sg",
    107: "Bh",
    108: "Hs",
    109: "Mt",
    110: "Ds",
    111: "Rg",
    112: "Cn",
    113: "Nh",
    114: "Fl",
    115: "Mc",
    116: "Lv",
    117: "Ts",
    118: "Og",
}
SYM_DICT = dict((v, k) for k, v in Z_DICT.items())


def Z_to_elem(Z: int) -> str:
    """
    Converts atomic number to element symbol.

    Parameters
    ----------
    Z : int
        Atomic number.

    Returns
    -------
    str
        Element string.

    Examples
    --------
    >>> rd.utils.Z_to_elem(1)
    'H'
    >>> rd.utils.Z_to_elem(35)
    'Br'

    """

    return Z_DICT[Z]


def elem_to_Z(sym: str) -> int:
    """
    Converts element symbol to atomic number.

    Parameters
    ----------
    sym : str
        Element string.

    Returns
    -------
    int
        Atomic number.

    Examples
    --------
    >>> rd.utils.elem_to_Z('H')
    1
    >>> rd.utils.elem_to_Z('Br')
    35

    """

    return SYM_DICT[sym]


def build_id(Z: int, A: int, state: str = "") -> int:
    """
    Builds a canonical nuclide id from atomic number, atomic mass, and
    and energy state.

    Parameters
    ----------
    Z : int
        Atomic number.
    A : int
        Atomic mass.
    state : str
        energy state.

    Returns
    -------
    int
        Canonical nuclide id.

    Examples
    --------
    >>> rd.utils.build_id(1,2)
    10020000
    >>> rd.utils.build_id(28,56,'m')
    280560001

    """

    if state != "":
        if state == "m":
            state_int = 1
        elif state == "n":
            state_int = 2
        else:
            raise ValueError(state + " is not a valid energy state.")
    else:
        state_int = 0

    canonical_id = (Z * 10000000) + (A * 10000) + state_int

    return canonical_id


def build_nuclide_string(Z: int, A: int, meta_state: str = "") -> str:
    """
    Builds a nuclide string from given atomic mass and number.

    Parameters
    ----------
    Z : int
        Atomic number.
    A : int
        Atomic mass.
    meta_state : str, optional
        Metastable state indicator character, ex. 'm' for the first
        atomic metastable state.

    Returns
    -------
    str
        Nuclide string built in symbol - mass number format.

    Examples
    --------
    >>> rd.utils.build_nuclide_string(26, 56)
    'Fe-56'
    >>> rd.utils.build_nuclide_string(28, 56, 'm')
    'Fe-56m'

    """

    if Z not in Z_DICT.keys():
        raise ValueError(str(Z) + " is not a valid atomic number")

    return_string = Z_DICT[Z] + "-" + str(A) + meta_state

    return return_string


def parse_nuclide_str(nuclide: str) -> str:
    """
    Parses a nuclide string from e.g. '241Pu' or 'Pu241' format to 'Pu-241' format. Note this
    function works for both radioactive and stable nuclides.

    Parameters
    ----------
    nuclide : str
        Nuclide string.

    Returns
    -------
    str
        Nuclide string parsed in symbol - mass number format.

    Examples
    --------
    >>> rd.utils.parse_nuclide_str('222Rn')
    'Rn-222'
    >>> rd.utils.parse_nuclide_str('Ca40')
    'Ca-40'

    """

    letter_flag, number_flag = False, False
    for char in nuclide:
        if char.isalpha():
            letter_flag = True
        if char.isdigit():
            number_flag = True
        if letter_flag and number_flag:
            break

    if not (letter_flag and number_flag) or len(nuclide) < 2 or len(nuclide) > 7:
        raise ValueError(str(nuclide) + " is not a valid nuclide string.")

    while nuclide[0].isdigit():  # Re-order inputs e.g. 99mTc to Tc99m.
        nuclide = nuclide[1:] + nuclide[0]
    if nuclide[0] in ["m", "n"]:
        nuclide = nuclide[1:] + nuclide[0]

    for i in range(1, len(nuclide)):  # Add hyphen e.g. Tc99m to Tc-99m.
        if nuclide[i].isdigit():
            if nuclide[i - 1] != "-":
                nuclide = nuclide[:i] + "-" + nuclide[i:]
            break

    return nuclide


def parse_id(input_id: int) -> str:
    """
    Parses a nuclide canonical id in zzzaaammmm format into symbol -
    mass number format.

    Parameters
    ----------
    input_id : int
        Nuclide in canonical id format.

    Returns
    -------
    str
        Nuclide string parsed in symbol - mass number format.

    Examples
    --------
    >>> rd.utils.parse_id(190400000)
    'K-40'
    >>> rd.utils.parse_id(280560001)
    'Ni-56m'

    """

    id_zzzaaa = int(input_id / 10000)
    state_digits = input_id - (id_zzzaaa * 10000)
    state = ""
    if state_digits == 1:
        state = "m"
    elif state_digits == 2:
        state = "n"
    Z = int(id_zzzaaa / 1000)
    A = id_zzzaaa - (Z * 1000)
    name = build_nuclide_string(Z, A, state)

    return name


def parse_nuclide(
    input_nuclide: Union[str, int], nuclides: List[str], dataset_name: str
) -> str:
    """
    Parses a nuclide string or canonical id into symbol - mass number
    format and checks whether the nuclide is contained in the decay
    dataset.

    Parameters
    ----------
    input_nuclide : str or int
        Nuclide name string or canonical id in zzzaaammmm format.
    nuclides : List[str]
        List of all the nuclides in the decay dataset.
    dataset_name : str
        Name of the decay dataset.

    Returns
    -------
    str
        Nuclide string parsed in symbol - mass number format.

    Raises
    ------
    ValueError
        If the input nuclide string or id is invalid or the nuclide is
        not contained in the decay dataset.
    TypeError
        If the input is an invalid type, a string or integer is
        expected.

    Examples
    --------
    >>> rd.utils.parse_nuclide('222Rn', rd.DEFAULTDATA.nuclides, rd.DEFAULTDATA.dataset_name)
    'Rn-222'
    >>> rd.utils.parse_nuclide('Ba137m', rd.DEFAULTDATA.nuclides, rd.DEFAULTDATA.dataset_name)
    'Ba-137m'
    >>> rd.utils.parse_nuclide(280560001, rd.DEFAULTDATA.nuclides, rd.DEFAULTDATA.dataset_name)
    'Ni-56m'

    """

    original_input = input_nuclide
    if isinstance(input_nuclide, int):
        name = parse_id(input_nuclide)
    elif isinstance(input_nuclide, str):
        name = input_nuclide
    else:
        raise TypeError("Invalid input type, expected int or str")
    nuclide = parse_nuclide_str(name)

    if nuclide not in nuclides:
        raise ValueError(
            str(original_input)
            + " is not a valid nuclide in "
            + dataset_name
            + " decay dataset."
        )

    return nuclide


def add_dictionaries(
    dict_a: Dict[str, Union[float, Expr]], dict_b: Dict[str, Union[float, Expr]]
) -> Dict[str, Union[float, Expr]]:
    """
    Adds together two dictionaries of nuclide strings and associated quantities. Supports both
    floats or SymPy quantities.

    Parameters
    ----------
    dict_a : dict
        First dictionary containing nuclide strings as keys and quantitites as values.
    dict_b : dict
        Second dictionary containing nuclide strings as keys and quantitites as values.

    Returns
    -------
    dict
        Combined dictionary containing the nuclides in both dict_a and dict_b, where the
        quantities have been added together when a nuclide is present in both input
        dictionaries.

    Examples
    --------
    >>> dict_a = {'Pm-141': 1.0, 'Rb-78': 2.0}
    >>> dict_b = {'Pm-141': 3.0, 'Rb-90': 4.0}
    >>> rd.utils.add_dictionaries(dict_a, dict_b)
    {'Pm-141': 4.0, 'Rb-78': 2.0, 'Rb-90': 4.0}

    """

    new_dict = dict_a.copy()
    for nuclide, quantity in dict_b.items():
        if nuclide in new_dict:
            new_dict[nuclide] = new_dict[nuclide] + quantity
        else:
            new_dict[nuclide] = quantity

    return new_dict


def sort_dictionary_alphabetically(
    input_inv_dict: Dict[str, Union[float, Expr]]
) -> Dict[str, Union[float, Expr]]:
    """
    Sorts a dictionary alphabetically by its keys.

    Parameters
    ----------
    input_inv_dict : dict
        Dictionary containing nuclide strings or Nuclide objects as keys and numbers
        as values.

    Returns
    -------
    dict
        Inventory dictionary which has been sorted by the nuclides alphabetically.

    Examples
    --------
    >>> rd.utils.sort_dictionary_alphabetically({'U-235': 1.2, 'Tc-99m': 2.3, 'Tc-99': 5.8})
    {'Tc-99': 5.8, 'Tc-99m': 2.3, 'U-235': 1.2}

    """

    return dict(sorted(input_inv_dict.items(), key=lambda x: x[0]))


def sort_list_according_to_dataset(
    input_list: List[str], key_dict: Dict[str, int]
) -> List[str]:
    """
    Sorts a list of nuclides based on their order of appearence in the decay dataset.

    Parameters
    ----------
    input_list : list
        List of nuclide strings to be sorted.
    key_dict : dict
        Dictionary from the decay dataset with nuclide strings as keys and their position
        (integers) in the decay dataset.

    Returns
    -------
    list
        Sorted nuclide list.

    Examples
    --------
    >>> rd.utils.sort_list_according_to_dataset(['Tc-99', 'Tc-99m'], rd.DEFAULTDATA.nuclide_dict)
    ['Tc-99m', 'Tc-99']

    """

    return sorted(input_list, key=lambda nuclide: key_dict[nuclide])



"""
Unit tests for utils.py functions.
"""

import unittest
from sympy import Integer, log
from radioactivedecay.utils import (
    Z_to_elem,
    elem_to_Z,
    build_id,
    build_nuclide_string,
    parse_id,
    parse_nuclide_str,
    parse_nuclide,
    add_dictionaries,
    sort_dictionary_alphabetically,
    sort_list_according_to_dataset,
)


class TestUtilsFunctions(unittest.TestCase):
    """
    Unit tests for the utils.py functions.
    """

    def test_Z_to_elem(self) -> None:
        """
        Test the conversion of atomic number to element symbol.
        """

        self.assertEqual(Z_to_elem(1), "H")
        self.assertEqual(Z_to_elem(20), "Ca")
        self.assertEqual(Z_to_elem(26), "Fe")

    def test_elem_to_Z(self) -> None:
        """
        Test the conversion of element symbol to atomic number.
        """

        self.assertEqual(elem_to_Z("H"), 1)
        self.assertEqual(elem_to_Z("Ca"), 20)
        self.assertEqual(elem_to_Z("Fe"), 26)

    def test_built_id(self) -> None:
        """
        Test the canonical id builder.
        """

        self.assertEqual(build_id(26, 56), 260560000)
        self.assertEqual(build_id(53, 118), 531180000)
        self.assertEqual(build_id(53, 118, "m"), 531180001)
        self.assertEqual(build_id(65, 156, "n"), 651560002)

        with self.assertRaises(ValueError):
            build_id(65, 156, "z")

    def test_built_nuclide_string(self) -> None:
        """
        Test the nuclide string builder.
        """

        self.assertEqual(build_nuclide_string(26, 56), "Fe-56")
        self.assertEqual(build_nuclide_string(53, 118), "I-118")
        self.assertEqual(build_nuclide_string(53, 118, "m"), "I-118m")
        self.assertEqual(build_nuclide_string(65, 156, "n"), "Tb-156n")

        with self.assertRaises(ValueError):
            build_nuclide_string(999, 1000, "z")

    def test_parse_id(self) -> None:
        """
        Test the canonical id to nuclide string converter.
        """

        self.assertEqual(parse_id(260560000), "Fe-56")
        self.assertEqual(parse_id(531180000), "I-118")
        self.assertEqual(parse_id(531180001), "I-118m")
        self.assertEqual(parse_id(651560002), "Tb-156n")

    def test_parse_nuclide_str(self) -> None:
        """
        Test the parsing of nuclide strings.
        """

        self.assertEqual(parse_nuclide_str("Ca-40"), "Ca-40")
        self.assertEqual(parse_nuclide_str("Ca40"), "Ca-40")
        self.assertEqual(parse_nuclide_str("40Ca"), "Ca-40")

    def test_parse_nuclide(self) -> None:
        """
        Test the parsing of nuclide strings.
        """

        nuclides = [
            "H-3",
            "Be-7",
            "C-10",
            "Ne-19",
            "I-118",
            "Pd-100",
            "Cl-34m",
            "I-118m",
            "Tb-156m",
            "Tb-156n",
        ]
        dataset_name = "test"

        # Re-formatting of acceptable strings e.g. 100Pd -> Pd-100
        self.assertEqual(parse_nuclide("H-3", nuclides, dataset_name), "H-3")
        self.assertEqual(parse_nuclide("H3", nuclides, dataset_name), "H-3")
        self.assertEqual(parse_nuclide("3H", nuclides, dataset_name), "H-3")
        self.assertEqual(parse_nuclide(10030000, nuclides, dataset_name), "H-3")
        self.assertEqual(parse_nuclide("Be-7", nuclides, dataset_name), "Be-7")
        self.assertEqual(parse_nuclide("Be7", nuclides, dataset_name), "Be-7")
        self.assertEqual(parse_nuclide("7Be", nuclides, dataset_name), "Be-7")
        self.assertEqual(parse_nuclide(40070000, nuclides, dataset_name), "Be-7")
        self.assertEqual(parse_nuclide("C-10", nuclides, dataset_name), "C-10")
        self.assertEqual(parse_nuclide("C10", nuclides, dataset_name), "C-10")
        self.assertEqual(parse_nuclide("10C", nuclides, dataset_name), "C-10")
        self.assertEqual(parse_nuclide(60100000, nuclides, dataset_name), "C-10")
        self.assertEqual(parse_nuclide("Ne-19", nuclides, dataset_name), "Ne-19")
        self.assertEqual(parse_nuclide("Ne19", nuclides, dataset_name), "Ne-19")
        self.assertEqual(parse_nuclide("19Ne", nuclides, dataset_name), "Ne-19")
        self.assertEqual(parse_nuclide(100190000, nuclides, dataset_name), "Ne-19")
        self.assertEqual(parse_nuclide("I-118", nuclides, dataset_name), "I-118")
        self.assertEqual(parse_nuclide("I118", nuclides, dataset_name), "I-118")
        self.assertEqual(parse_nuclide("118I", nuclides, dataset_name), "I-118")
        self.assertEqual(parse_nuclide(531180000, nuclides, dataset_name), "I-118")
        self.assertEqual(parse_nuclide("Pd-100", nuclides, dataset_name), "Pd-100")
        self.assertEqual(parse_nuclide("Pd100", nuclides, dataset_name), "Pd-100")
        self.assertEqual(parse_nuclide("100Pd", nuclides, dataset_name), "Pd-100")
        self.assertEqual(parse_nuclide(461000000, nuclides, dataset_name), "Pd-100")
        self.assertEqual(parse_nuclide("Cl-34m", nuclides, dataset_name), "Cl-34m")
        self.assertEqual(parse_nuclide("Cl34m", nuclides, dataset_name), "Cl-34m")
        self.assertEqual(parse_nuclide("34mCl", nuclides, dataset_name), "Cl-34m")
        self.assertEqual(parse_nuclide(170340001, nuclides, dataset_name), "Cl-34m")
        self.assertEqual(parse_nuclide("I-118m", nuclides, dataset_name), "I-118m")
        self.assertEqual(parse_nuclide("I118m", nuclides, dataset_name), "I-118m")
        self.assertEqual(parse_nuclide("118mI", nuclides, dataset_name), "I-118m")
        self.assertEqual(parse_nuclide(531180001, nuclides, dataset_name), "I-118m")
        self.assertEqual(parse_nuclide("Tb-156m", nuclides, dataset_name), "Tb-156m")
        self.assertEqual(parse_nuclide("Tb156m", nuclides, dataset_name), "Tb-156m")
        self.assertEqual(parse_nuclide("156mTb", nuclides, dataset_name), "Tb-156m")
        self.assertEqual(parse_nuclide(651560001, nuclides, dataset_name), "Tb-156m")
        self.assertEqual(parse_nuclide("Tb-156n", nuclides, dataset_name), "Tb-156n")
        self.assertEqual(parse_nuclide("Tb156n", nuclides, dataset_name), "Tb-156n")
        self.assertEqual(parse_nuclide("156nTb", nuclides, dataset_name), "Tb-156n")
        self.assertEqual(parse_nuclide(651560002, nuclides, dataset_name), "Tb-156n")

        # Catch erroneous strings
        with self.assertRaises(TypeError):
            parse_nuclide(1.2, nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("H", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("A1", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("1A", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("H-4", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("H4", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("4H", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("Pb-198m", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("Pbo-198m", nuclides, dataset_name)

    def test_add_dictionaries(self) -> None:
        """
        Test function which adds two inventory dictionaries together.
        """

        dict1 = {"Pm-141": 1.0, "Rb-78": 2.0}
        dict2 = {"Pm-141": 3.0, "Rb-90": 4.0}
        self.assertEqual(
            add_dictionaries(dict1, dict2),
            {"Pm-141": 4.0, "Rb-78": 2.0, "Rb-90": 4.0},
        )

        dict1 = {"Pm-141": Integer(2) * log(3), "Rb-78": Integer(4) / log(5)}
        dict2 = {"Pm-141": log(3) / Integer(7), "Rb-90": Integer(9)}
        self.assertEqual(
            add_dictionaries(dict1, dict2),
            {
                "Pm-141": Integer(15) * log(3) / Integer(7),
                "Rb-78": Integer(4) / log(5),
                "Rb-90": Integer(9),
            },
        )

    def test_sort_dictionary_alphabetically(self) -> None:
        """
        Test the sorting of a dictionary by its keys alphabetically.
        """

        inv_dict = {"U-235": 1.2, "Tc-99m": 2.3, "Tc-99": 5.8}
        self.assertEqual(
            sort_dictionary_alphabetically(inv_dict),
            {"Tc-99": 5.8, "Tc-99m": 2.3, "U-235": 1.2},
        )

        inv_dict = {"U-235": Integer(1), "Tc-99m": Integer(2), "Tc-99": Integer(3)}
        self.assertEqual(
            sort_dictionary_alphabetically(inv_dict),
            {"Tc-99": Integer(3), "Tc-99m": Integer(2), "U-235": Integer(1)},
        )

    def test_sort_list_according_to_dataset(self) -> None:
        """
        Test the sorting of list of nuclides according to their position in the decay dataset.
        """

        nuclide_list = ["Tc-99", "Tc-99m"]
        nuclide_dict = {"Tc-99m": 0, "Tc-99": 1}
        self.assertEqual(
            sort_list_according_to_dataset(nuclide_list, nuclide_dict),
            ["Tc-99m", "Tc-99"],
        )


if __name__ == "__main__":
    unittest.main()



"""
The nuclide module defines the ``Nuclide`` class. Each ``Nuclide``
instance represents one nuclide, and can be contructed from a nuclide
string or `zzzaaassss` canonical id. The class properties provide an
easy way of extracting atomic number and mass information, as well as a
clean name string. Furthermore, additional methods provide an access
point for mass data of nuclides, and the decay data of radionuclides,
if present in a specified dataset. The default decay dataset used if
none is supplied to the constructor is rd.DEFAULTDATA.

The code examples shown in the docstrings assume the
``radioactivedecay`` package has been imported as:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

"""

from collections import deque
from typing import Any, Dict, List, Tuple, Union, Optional
import matplotlib
import networkx as nx
import numpy as np
from radioactivedecay.decaydata import DecayData, DEFAULTDATA
from radioactivedecay.plots import (
    _parse_nuclide_label,
    _parse_decay_mode_label,
    _check_fig_axes,
)
from radioactivedecay.utils import parse_nuclide, elem_to_Z, build_id


class Nuclide:
    """
    ``Nuclide`` instances serve as name and atomic number/mass parsing
    objects for any nuclide or element.

    Parameters
    ----------
    input_nuclide : str or int
        Input value for instantiation. Can be nuclide string in name
        format (with or without hyphen), or canonical id (zzzaaassss).
    decay_data : DecayData, optional
        Decay dataset (default is the ICRP-107 dataset).

    Attributes
    ----------
    nuclide: str
        Nuclide name string.
    Z : int
        Atomic number.
    A : int
        Atomic mass number.
    id : int
        Canonical nuclide id, in zzzaaassss form. Ground state is 0000,
        first excited state ("m") is 0001, second ("n") is 0002, etc.
    decay_data : DecayData
        Decay dataset.
    atomic_mass : float
        Atomic weight of the nuclide, in g/mol.
    prog_bf_mode : int
        Dictionary containing direct progeny as keys, and a list
        containing the branching fraction and the decay mode for that
        progeny as values.

    Examples
    --------
    >>> rd.Nuclide('K-40')
    Nuclide: K-40
    >>> rd.Nuclide('K40')
    Nuclide: K-40
    >>> rd.Nuclide(190400000)
    Nuclide: K-40
    >>> rd.Nuclide(280560001)
    Nuclide: Ni-56m

    """

    def __init__(
        self, input_nuclide: Union[str, int], decay_data: DecayData = DEFAULTDATA
    ) -> None:
        self.decay_data = decay_data
        self.parse_name(input_nuclide)
        index = decay_data.nuclide_dict[self.nuclide]
        mass = decay_data.scipy_data.atomic_masses[index]
        self.atomic_mass = mass
        self.prog_bf_mode: Dict[
            str, List[Union[float, str]]
        ] = decay_data.prog_bfs_modes[decay_data.nuclide_dict[self.nuclide]]

    def parse_name(self, input_nuclide: Union[str, int]) -> None:
        """
        Parse input and set atomic data attributes.
        """

        self.nuclide = parse_nuclide(
            input_nuclide, self.decay_data.nuclides, self.decay_data.dataset_name
        )
        self.Z = elem_to_Z(self.nuclide.split("-")[0])
        self.A = int(self.nuclide.split("-")[1].strip("mn"))
        self.state = self.nuclide.split("-")[1].strip("0123456789")
        self.id = build_id(self.Z, self.A, self.state)

    def __repr__(self) -> str:
        rep = (
            "Nuclide: "
            + self.nuclide
            + ", decay dataset: "
            + self.decay_data.dataset_name
        )

        return rep

    def __eq__(self, other: object) -> bool:
        """
        Check whether two ``Radionuclide`` instances are equal with ``==`` operator.
        """

        if not isinstance(other, Nuclide):
            return NotImplemented
        return self.nuclide == other.nuclide and self.decay_data == other.decay_data

    def __ne__(self, other: object) -> bool:
        """
        Check whether two ``Radionuclide`` instances are not equal with ``!=`` operator.
        """

        if not isinstance(other, Nuclide):
            return NotImplemented
        return not self.__eq__(other)

    def __hash__(self) -> int:
        """
        Hash function for ``Radionuclide`` instances.
        """

        return hash((self.nuclide, self.decay_data.dataset_name))

    def half_life(self, units: str = "s") -> Union[float, str]:
        """
        Returns the half-life of a nuclide as a float in your chosen
        units, or as a human-readable string with appropriate units.

        Parameters
        ----------
        units : str, optional
            Units for half-life. Options are 'ps', 'ns', 'μs', 'us',
            'ms', 's', 'm', 'h', 'd', 'y', 'ky', 'My', 'By', 'Gy',
            'Ty', 'Py', and common spelling variations. Default is 's',
            i.e. seconds. Use 'readable' to get a string of the
            half-life in human-readable units.

        Returns
        -------
        float or str
            Radionuclide half-life.

        Examples
        --------
        >>> K40 = rd.Nuclide('K-40')
        >>> K40.half_life('y')
        1251000000.0
        >>> Fe56 = rd.Nuclide('Fe-56')
        >>> Fe56.half_life('readable')
        'stable'

        """

        return self.decay_data.half_life(self.nuclide, units)

    def progeny(self) -> List[str]:
        """
        Returns the direct progeny of a radionuclide.

        Returns
        -------
        list
            List of the direct progeny of the radionuclide, ordered by
            decreasing branching fraction.

        Examples
        --------
        >>> K40 = rd.Nuclide('K-40')
        >>> K40.progeny()
        ['Ca-40', 'Ar-40']

        """

        return list(self.prog_bf_mode.keys())

    def branching_fractions(self) -> List[float]:
        """
        Returns the branching fractions to the direct progeny of a
        radionuclide.

        Returns
        -------
        list
            List of branching fractions.

        Examples
        --------
        >>> K40 = rd.Nuclide('K-40')
        >>> K40.branching_fractions()
        [0.8914, 0.1086]

        """

        branching_fractions: List[float] = [
            bf_mode[0] for bf_mode in list(self.prog_bf_mode.values())
        ]
        return branching_fractions

    def decay_modes(self) -> List[str]:
        """
        Returns the decay modes for a radionuclide, as defined in the
        decay dataset. Note: the decay mode strings returned are not
        lists of all the different radiation types emitted during the
        parent to progeny decay processes. They are the labels defined
        in the decay dataset to classify the parent to progeny decay
        type (e.g. '\u03b1', '\u03b2-' or 'IT').

        Returns
        -------
        list
            List of decay modes.

        Examples
        --------
        >>> K40 = rd.Nuclide('K-40')
        >>> K40.decay_modes()
        ['\u03b2-', '\u03b2+ & EC']

        """

        decay_modes: List[str] = [
            bf_mode[1] for bf_mode in list(self.prog_bf_mode.values())
        ]
        return decay_modes

    def plot(
        self,
        label_pos: float = 0.5,
        fig: Optional[matplotlib.figure.Figure] = None,
        axes: Optional[matplotlib.axes.Axes] = None,
        kwargs_draw: Optional[Dict[str, Any]] = None,
        kwargs_edge_labels: Optional[Dict[str, Any]] = None,
    ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        """
        Plots a diagram of the decay chain of a radionuclide. Then
        creates a NetworkX DiGraph and plot of it using NetworkX's
        Matplotlib-based plotting functionality.

        Some of the NetworkX default plotting parameters are changed to
        produce nice decay chain diagrams. However, users retain
        control over these parameters via kwargs_draw and
        kwargs_edge_labels. For more information on the various
        NetworkX plotting parameters, refer to its `documentation
        <https://networkx.org/documentation/stable/reference/drawing.html>`_.

        Parameters
        ----------
        label_pos : float, optional
            Position of labels along edges. Default is 0.5. If you find
            that edge labels are overlapping in the decay chain
            diagram, try increasing this parameter to e.g. 0.66.
        fig : None or matplotlib.figure.Figure, optional
            matplotlib figure object to use, or None makes
            ``radioactivedecay`` create one (default is None).
        axes : None or matplotlib.axes.Axes, optional
            matplotlib axes object to use, or None makes
            ``radioactivedecay`` create one (default is None).
        **kwargs_draw, optional
            Keyword arguments for networkx.draw().
        **kwargs_edge_labels, optional
            Keyword arguments for networkx.draw_networkx_edge_labels().

        Returns
        -------
        fig : matplotlib.figure.Figure
            matplotlib figure object used to plot the decay chain.
        ax : matplotlib.axes.Axes
            matplotlib axes object used to plot the decay chain.

        """

        digraph, max_generation, max_xpos = _build_decay_digraph(self, nx.DiGraph())

        positions = nx.get_node_attributes(digraph, "pos")
        node_labels = nx.get_node_attributes(digraph, "label")
        edge_labels = nx.get_edge_attributes(digraph, "label")

        fig, axes = _check_fig_axes(
            fig, axes, figsize=(3 * max_xpos + 1.5, 3 * max_generation + 1.5)
        )

        if kwargs_draw is None:
            kwargs_draw = {}
        if "node_size" not in kwargs_draw:
            kwargs_draw["node_size"] = 6000
        if "node_color" not in kwargs_draw:
            kwargs_draw["node_color"] = "#FFFFFF"
        if "edgecolors" not in kwargs_draw:
            kwargs_draw["edgecolors"] = "#000000"

        nx.draw(
            G=digraph,
            pos=positions,
            ax=axes,
            labels=node_labels,
            **kwargs_draw,
        )

        if kwargs_edge_labels is None:
            kwargs_edge_labels = {}
        if "font_size" not in kwargs_edge_labels:
            kwargs_edge_labels["font_size"] = 12
        if "bbox" not in kwargs_edge_labels:
            kwargs_edge_labels["bbox"] = dict(
                boxstyle=None, ec=(1.0, 1.0, 1.0), fc=(1.0, 1.0, 1.0)
            )
        if "rotate" not in kwargs_edge_labels:
            kwargs_edge_labels["rotate"] = False

        nx.draw_networkx_edge_labels(
            G=digraph,
            pos=positions,
            edge_labels=edge_labels,
            label_pos=label_pos,
            ax=axes,
            **kwargs_edge_labels,
        )

        axes.set_xlim(-0.3, max_xpos + 0.3)
        axes.set_ylim(-max_generation - 0.3, 0.3)

        return fig, axes


def _build_decay_digraph(
    parent: Nuclide,
    digraph: nx.classes.digraph.DiGraph,
) -> nx.classes.digraph.DiGraph:
    """
    Build a networkx DiGraph for the decay chain of this nuclide.

    Parameters
    ----------
    parent : Radionuclide
        Radionuclide instance of the parent nuclide of the decay chain.
    digraph : networkx.classes.digraph.DiGraph
        DiGraph for the decay chain.

    Returns
    -------
    digraph : networkx.classes.digraph.DiGraph
        DiGraph of the decay chain.
    max_generation : int
        Number of generations of progeny in the decay chain.
    max_xpos : int
        Maximum number of progeny within any one generation of the decay chain.

    """

    generation_max_xpos = {0: 0}

    dequeue = deque([parent.nuclide])
    generations = deque([0])
    xpositions = deque([0])
    node_label = (
        _parse_nuclide_label(parent.nuclide) + "\n" + str(parent.half_life("readable"))
    )
    digraph.add_node(parent.nuclide, generation=0, xpos=0, label=node_label)
    seen = {parent.nuclide}

    while len(dequeue) > 0:
        parent_name = dequeue.popleft()
        generation = generations.popleft() + 1
        xpos = xpositions.popleft()
        if generation not in generation_max_xpos:
            generation_max_xpos[generation] = -1
        parent = Nuclide(parent_name, parent.decay_data)

        progeny = parent.progeny()
        branching_fractions = parent.branching_fractions()
        decay_modes = parent.decay_modes()

        if xpos < generation_max_xpos[generation] + 1:
            xpos = generation_max_xpos[generation] + 1
        xcounter = 0
        for i, prog in enumerate(progeny):
            if prog not in seen:
                node_label = _parse_nuclide_label(prog)
                if prog in parent.decay_data.nuclide_dict:
                    node_label += "\n" + str(
                        parent.decay_data.half_life(prog, "readable")
                    )
                    if np.isfinite(parent.decay_data.half_life(prog)):
                        dequeue.append(prog)
                        generations.append(generation)
                        xpositions.append(xpos + xcounter)
                if prog == "SF":
                    prog = parent.nuclide + "_SF"

                digraph.add_node(
                    prog,
                    generation=generation,
                    xpos=xpos + xcounter,
                    label=node_label,
                )
                seen.add(prog)

                if xpos + xcounter > generation_max_xpos[generation]:
                    generation_max_xpos[generation] = xpos + xcounter
                xcounter += 1

            edge_label = (
                _parse_decay_mode_label(decay_modes[i])
                + "\n"
                + str(branching_fractions[i])
            )
            digraph.add_edge(parent.nuclide, prog, label=edge_label)

    for node in digraph:
        digraph.nodes[node]["pos"] = (
            digraph.nodes[node]["xpos"],
            digraph.nodes[node]["generation"] * -1,
        )

    return digraph, max(generation_max_xpos), max(generation_max_xpos.values())



"""
Unit tests for nuclide.py functions, classes and methods.
"""

import unittest
from radioactivedecay.decaydata import DecayData
from radioactivedecay.nuclide import Nuclide


class TestNuclide(unittest.TestCase):
    """
    Unit tests for the nuclide.py Nuclide class.
    """

    def test_nuclide_instantiation(self) -> None:
        """
        Test instantiation of Nuclide objects.
        """

        nuc = Nuclide("Rn-222")
        self.assertEqual(nuc.nuclide, "Rn-222")
        self.assertEqual(nuc.prog_bf_mode, {"Po-218": [1.0, "\u03b1"]})

        nuc = Nuclide("222Rn")
        self.assertEqual(nuc.nuclide, "Rn-222")
        self.assertEqual(nuc.prog_bf_mode, {"Po-218": [1.0, "\u03b1"]})

        nuc = Nuclide(611450000)
        self.assertEqual(nuc.nuclide, "Pm-145")
        self.assertEqual(
            nuc.prog_bf_mode, {"Nd-145": [1.0, "EC"], "Pr-141": [2.8e-09, "\u03b1"]}
        )

    def test_nuclide_Z(self) -> None:
        """
        Test Nuclide Z attribute.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(nuc.Z, 1)

    def test_nuclide_A(self) -> None:
        """
        Test Nuclide A attribute.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(nuc.A, 3)

    def test_nuclide_id(self) -> None:
        """
        Test Nuclide id attribute.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(nuc.id, 10030000)

        nuc = Nuclide("I-118m")
        self.assertEqual(nuc.id, 531180001)

        nuc = Nuclide(190400000)
        self.assertEqual(nuc.id, 190400000)

    def test_radionuclide_half_life(self) -> None:
        """
        Test Nuclide half_life() method.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(nuc.half_life(), 388781329.30560005)
        self.assertEqual(nuc.half_life("y"), 12.32)
        self.assertEqual(nuc.half_life("readable"), "12.32 y")

    def test_radionuclide_progeny(self) -> None:
        """
        Test Nuclide half_life() method.
        """

        nuc = Nuclide("K-40")
        self.assertEqual(nuc.progeny()[0], "Ca-40")
        self.assertEqual(nuc.progeny()[1], "Ar-40")

    def test_radionuclide_branching_fractions(self) -> None:
        """
        Test Nuclide branching_fractions() method.
        """

        nuc = Nuclide("K-40")
        self.assertEqual(nuc.branching_fractions()[0], 0.8914)
        self.assertEqual(nuc.branching_fractions()[1], 0.1086)

    def test_radionuclide_decay_modes(self) -> None:
        """
        Test Nuclide decay_modes() method.
        """

        nuc = Nuclide("K-40")
        self.assertEqual(nuc.decay_modes()[0], "\u03b2-")
        self.assertEqual(nuc.decay_modes()[1], "\u03b2+ & EC")

    def test_radionuclide_plot(self) -> None:
        """
        Test Nuclide plot() method.

        Only testing auto-generation of limits so far.
        """

        nuc = Nuclide("H-3")
        _, axes = nuc.plot()
        self.assertEqual(axes.get_xlim(), (-0.3, 0.3))
        self.assertEqual(axes.get_ylim(), (-1.3, 0.3))

        nuc = Nuclide("Mo-99")
        _, axes = nuc.plot()
        self.assertEqual(axes.get_xlim(), (-0.3, 1.3))
        self.assertEqual(axes.get_ylim(), (-2.3, 0.3))

        nuc = Nuclide("Es-256")
        _, axes = nuc.plot()
        self.assertEqual(axes.get_xlim(), (-0.3, 2.3))
        self.assertEqual(axes.get_ylim(), (-19.3, 0.3))

    def test_radionuclide___repr__(self) -> None:
        """
        Test Nuclide __repr__ strings.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(
            nuc.__repr__(),
            "Nuclide: H-3, decay dataset: icrp107_ame2020_nubase2020",
        )

    def test_nuclide___eq__(self) -> None:
        """
        Test Nuclide equality.
        """

        nuc1 = Nuclide("K-40")
        nuc2 = Nuclide("40K")
        nuc3 = Nuclide(190400000)
        self.assertEqual(nuc1, nuc2)
        self.assertEqual(nuc1, nuc3)

        decay_data = DecayData("icrp107_ame2020_nubase2020", load_sympy=True)
        nuc2 = Nuclide("K-40", decay_data)
        self.assertEqual(nuc1, nuc2)

        self.assertFalse(nuc1 == "random object")

    def test_radionuclide___ne__(self) -> None:
        """
        Test Nuclide inequality.
        """

        nuc1 = Nuclide("K-40")
        nuc2 = Nuclide("H-3")
        self.assertNotEqual(nuc1, nuc2)

        self.assertTrue(nuc1 != "random object")

    def test_radionuclide___hash__(self) -> None:
        """
        Test Nuclide hash function.
        """

        nuc = Nuclide("K-40")
        decay_data = DecayData("icrp107_ame2020_nubase2020", load_sympy=True)
        self.assertEqual(hash(nuc), hash(("K-40", decay_data.dataset_name)))


if __name__ == "__main__":
    unittest.main()
